import warnings
import itertools
from math import sqrt
import random
import xml.etree.ElementTree as ET
from abc import ABC, abstractmethod
from functools import reduce
from operator import mul
from typing import (
    Callable,
    Collection,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    Any,
)

import numpy as np

from negmas.common import AgentMechanismInterface, NamedObject
from negmas.generics import ienumerate, ivalues
from negmas.helpers import Distribution, PATH, ikeys, snake_case, get_full_type_name
from negmas.outcomes import (
    Issue,
    Outcome,
    outcome_as,
    outcome_as_dict,
    outcome_as_tuple,
    outcome_is_valid,
    sample_outcomes,
)
from negmas.serialization import serialize, deserialize

__all__ = [
    "UtilityDistribution",
    "UtilityValue",
    "UtilityDistribution",
    "UtilityFunction",
    "INVALID_UTILITY",
    "OutcomeUtilityMapping",
    "OutcomeUtilityMappings",
    "ExactUtilityValue",
]

INVALID_UTILITY = float("-inf")


# Helper Types just used to make type hinting more readable
OutcomeUtilityMapping = Union[
    Callable[[Union["Outcome", int, str, float]], "UtilityValue"],
    Mapping[Union[Sequence, Mapping, int, str, float], "UtilityValue"],
]
"""A mapping from an outcome to its utility value"""


OutcomeUtilityMappings = List[OutcomeUtilityMapping]
"""Maps from multi-issue or single-issue outcomes to Negotiator values."""


IssueUtilityFunctionMapping = Union[
    Callable[["Issue"], "UtilityFunction"], Mapping["Issue", "UtilityFunction"],
]
"""A mapping from issues to utility functions"""


IssueUtilityFunctionMappings = List[IssueUtilityFunctionMapping]
"""A list of `IssueUtilityFunctionMapping`"""


UtilityDistribution = Distribution
"""A probability distribution over utility values"""


class ExactUtilityValue(float):
    """Encapsulates a single offerable_outcomes utility_function value."""


UtilityValue = Union[UtilityDistribution, float]
"""
Either a utility_function distribution or an exact offerable_outcomes
utility_function value.

`UtilityFunction`s always return a `UtilityValue` which makes it easier to
implement algorithms relying  on probabilistic modeling of utility functions.
"""


class UtilityFunction(ABC, NamedObject):
    """The abstract base class for all utility functions.

    A utility function encapsulates a mapping from outcomes to UtilityValue(s).
    This is a generalization of standard
    utility functions that are expected to always return a real-value.
    This generalization is useful for modeling cases
    in which only partial knowledge of the utility function is available.

    To define a new utility function, you have to override one of the following
    methods:

        - `eval` to define a standard ufun mapping outcomes to utility values.
        - `is_better` to define prferences by a partial ordering over outcomes
          implemented through bilateral comparisons
        - `rank` to define the preferences by partial ordering defined by
          a ranking of outcomes.

    Args:
        name: Name of the utility function. If None, a random name will
              be given.
        reserved_value: The value to return if the input offer to
                        `__call__` is None
        outcome_type: The type to use when evauating utilities.
                      It can be tuple, dict, or any `OutcomeType`
        issue_names: The names of issues. Only needed if outcome_type is not tuple
        issues: The list of issues for which the ufun is defined (optional)
        ami: The `AgentMechanismInterface` for a mechanism for which the ufun
             is defined (optinoal)
        id: An optional system-wide unique identifier. You should not change
            the default value except in special circumstances like during
            serialization and should always guarantee system-wide uniquness
            if you set this value explicitly

    Remarks:
        - If ami is given, it overrides outcome_type, issues and issue_names
        - If issues is given, it overrides issue_names
        - One of `eval`, `is_better`, `rank` **MUST** be overriden, otherwise
          calling any of them will lead to an infinite loop which is very hard
          to debug.

    """

    def __init__(
        self,
        name: Optional[str] = None,
        reserved_value: UtilityValue = float("-inf"),
        outcome_type: Optional[Type] = None,
        issue_names: Optional[List[str]] = None,
        issues: List["Issue"] = None,
        ami: AgentMechanismInterface = None,
        id: str = None,
    ) -> None:
        super().__init__(name, id=id)
        self.reserved_value = reserved_value
        self._ami = ami
        self._inverse_initialzed = False
        self._issues = issues if ami is None or ami.issues is None else ami.issues
        self._outcome_type = outcome_type if ami is None else ami.outcome_type
        self.issue_names = (
            issue_names if self._issues is None else [_.name for _ in self._issues]
        )
        if self._outcome_type is not None and (
            not issubclass(self._outcome_type, tuple) and self.issue_names is None
        ):
            raise ValueError(
                "You must specify issue_names because you are using a non-tuple outcome type"
            )

    @property
    def ami(self):
        return self._ami

    @ami.setter
    def ami(self, value: AgentMechanismInterface):
        self._ami = value

    @property
    def issues(self):
        return self._issues

    @issues.setter
    def issues(self, value: List[Issue]):
        self._issues = value
        self.issue_names = [_.name for _ in value] if value else value

    @property
    def outcome_type(self):
        if not hasattr(self, "_outcome_type"):
            return None
        return self._outcome_type

    @outcome_type.setter
    def outcome_type(self, value: Type):
        self._outcome_type = value

    @property
    def is_dynamic(self):
        """
        Whether the utility function can potentially depend on negotiation
        state (mechanism information).

        - If this property is `False`,  the ufun can safely be assumed to be
          static (not dependent on negotiation state).
        - If this property is `True`, the ufun may depend on negotiation state
          but it may also not depend on it.
        """
        # TODO: fix this. Now we always attach ami when the ufun joins a neg.
        #       so having no ami is not a good proxy for being dynamic
        return self.ami is None

    @classmethod
    def from_genius(cls, file_name: PATH, **kwargs):
        """Imports a utility function from a GENIUS XML file.

        Args:

            file_name (str): File name to import from

        Returns:

            A utility function object (depending on the input file)


        Examples:

            >>> from negmas.utilities import UtilityFunction
            >>> import pkg_resources
            >>> u, d = UtilityFunction.from_genius(file_name = pkg_resources.resource_filename('negmas'
            ...                                      , resource_name='tests/data/Laptop/Laptop-C-prof1.xml'))
            >>> u.__class__.__name__
            'LinearUtilityAggregationFunction'
            >>> u.reserved_value
            0.0
            >>> d
            1.0

        Remarks:
            See ``from_xml_str`` for all the parameters

        """
        with open(file_name, "r") as f:
            xml_str = f.read()
            return cls.from_xml_str(xml_str=xml_str, **kwargs)

    @classmethod
    def to_genius(
        cls, u: "UtilityFunction", issues: List[Issue], file_name: PATH, **kwargs
    ):
        """Exports a utility function to a GENIUS XML file.

        Args:

            file_name (str): File name to export to
            u: utility function
            issues: The issues being considered as defined in the domain

        Returns:

            None


        Examples:

            >>> from negmas.utilities import UtilityFunction
            >>> from negmas.inout import load_genius_domain
            >>> import pkg_resources
            >>> _, _, issues = load_genius_domain(domain_file_name=pkg_resources.resource_filename('negmas'
            ...                                             , resource_name='tests/data/Laptop/Laptop-C-domain.xml')
            ...             , keep_issue_names=False)
            >>> u, discount = UtilityFunction.from_genius(file_name=pkg_resources.resource_filename('negmas'
            ...                                             , resource_name='tests/data/Laptop/Laptop-C-prof1.xml')
            ...             , keep_issue_names=False)
            >>> UtilityFunction.to_genius(u=u, issues=issues, discount_factor=discount
            ...     , file_name = pkg_resources.resource_filename('negmas'
            ...                                             , resource_name='tests/data/LaptopConv/Laptop-C-prof1.xml'))

        Remarks:
            See ``to_xml_str`` for all the parameters

        """
        with open(file_name, "w") as f:
            f.write(cls.to_xml_str(u=u, issues=issues, **kwargs))

    @classmethod
    def to_xml_str(
        cls, u: "UtilityFunction", issues: List[Issue], discount_factor=None
    ) -> str:
        """Exports a utility function to a well formatted string"""
        if issues is not None:
            n_issues = len(issues)
        else:
            n_issues = 0
        output = (
            f'<utility_space type="any" number_of_issues="{n_issues}">\n'
            f'<objective index="1" etype="objective" type="objective" description="" name="any">\n'
        )

        output += u.xml(issues=issues)
        if "</objective>" not in output:
            output += "</objective>\n"
            if discount_factor is not None:
                output += f'<discount_factor value="{discount_factor}" />\n'
        if u.reserved_value != float("-inf") and "<reservation value" not in output:
            output += f'<reservation value="{u.reserved_value}" />\n'
        if "</utility_space>" not in output:
            output += "</utility_space>\n"
        return output

    @classmethod
    def from_xml_str(
        cls,
        xml_str: str,
        domain_issues: Optional[List[Issue]] = None,
        force_single_issue=False,
        force_numeric=False,
        keep_issue_names=True,
        keep_value_names=True,
        safe_parsing=True,
        normalize_utility=False,
        normalize_max_only=False,
        max_n_outcomes: int = 1_000_000,
        ignore_discount=False,
        ignore_reserved=False,
    ):
        """Imports a utility function from a GENIUS XML string.

        Args:

            xml_str (str): The string containing GENIUS style XML utility function definition
            domain_issues (List[Issue]): Optional issue space to confirm that the utility function is valid
            force_single_issue (bool): Tries to generate a MappingUtility function with a single issue which is the
            product of all issues in the input
            keep_issue_names (bool): Keep names of issues
            keep_value_names (bool): Keep names of values
            safe_parsing (bool): Turn on extra checks
            normalize_utility (bool): Normalize the output utilities to the range from 0 to 1
            normalize_max_only (bool): If True ensures that max(utility) = 1 but does not ensure that min(utility) = 0. and
                                      if false, ensures both max(utility) = 1 and min(utility) = 0
            max_n_outcomes (int): Maximum number of outcomes allowed (effective only if force_single_issue is True)

        Returns:

            A utility function object (depending on the input file)


        Examples:

            >>> import pkg_resources
            >>> u, _ = UtilityFunction.from_xml_str(open(pkg_resources.resource_filename('negmas'
            ...                                      , resource_name='tests/data/Laptop/Laptop-C-prof1.xml')
            ...                                      , 'r').read(), force_single_issue=False
            ...                                     , normalize_utility=True
            ...                                     , keep_issue_names=False, keep_value_names=True)

            >>> u, _ = UtilityFunction.from_xml_str(open(pkg_resources.resource_filename('negmas'
            ...                                      , resource_name='tests/data/Laptop/Laptop-C-prof1.xml')
            ...                                      , 'r').read()
            ...                                      , force_single_issue=True, normalize_utility=False)
            >>> assert abs(u(("Dell+60 Gb+19'' LCD",)) - 21.987727736172488) < 0.000001
            >>> assert abs(u(("HP+80 Gb+20'' LCD",)) - 22.68559475583014) < 0.000001

            >>> u, _ = UtilityFunction.from_xml_str(open(pkg_resources.resource_filename('negmas'
            ...                                      , resource_name='tests/data/Laptop/Laptop-C-prof1.xml')
            ...                                      , 'r').read(), force_single_issue=True
            ... , keep_issue_names=False, keep_value_names=False, normalize_utility=False)
            >>> assert abs(u((0,)) - 21.987727736172488) < 0.000001

            >>> u, _ = UtilityFunction.from_xml_str(open(pkg_resources.resource_filename('negmas'
            ...                                      , resource_name='tests/data/Laptop/Laptop-C-prof1.xml')
            ...                 , 'r').read(), force_single_issue=False, normalize_utility=False)
            >>> assert abs(u({'Laptop': 'Dell', 'Harddisk': '60 Gb', 'External Monitor': "19'' LCD"}) - 21.987727736172488) < 0.000001
            >>> assert abs(u({'Laptop': 'HP', 'Harddisk': '80 Gb', 'External Monitor': "20'' LCD"}) - 22.68559475583014) < 0.000001

            >>> u, _ = UtilityFunction.from_xml_str(open(pkg_resources.resource_filename('negmas'
            ...                                      , resource_name='tests/data/Laptop/Laptop-C-prof1.xml')
            ...                                      , 'r').read()
            ...                                      , force_single_issue=True, normalize_utility=True)

            >>> u, _ = UtilityFunction.from_xml_str(open(pkg_resources.resource_filename('negmas'
            ...                                      , resource_name='tests/data/Laptop/Laptop-C-prof1.xml')
            ...                                      , 'r').read(), force_single_issue=True
            ... , keep_issue_names=False, keep_value_names=False, normalize_utility=True)

            >>> u, _ = UtilityFunction.from_xml_str(open(pkg_resources.resource_filename('negmas'
            ...                                      , resource_name='tests/data/Laptop/Laptop-C-prof1.xml')
            ...         , 'r').read(), force_single_issue=False, normalize_utility=True)

        """
        from negmas.utilities.linear import LinearUtilityAggregationFunction
        from negmas.utilities.nonlinear import MappingUtilityFunction
        from negmas.utilities.complex import ComplexWeightedUtilityFunction
        from negmas.utilities.nonlinear import HyperRectangleUtilityFunction

        root = ET.fromstring(xml_str)
        if safe_parsing and root.tag != "utility_space":
            raise ValueError(f"Root tag is {root.tag}: Expected utility_space")

        if domain_issues is not None:
            if isinstance(domain_issues, list):
                domain_issues: Dict[str, Issue] = dict(
                    zip([_.name for _ in domain_issues], domain_issues)
                )
            elif isinstance(domain_issues, Issue) and force_single_issue:
                domain_issues = dict(zip([domain_issues.name], [domain_issues]))
        objective = None
        reserved_value = 0.0
        discount_factor = 0.0
        for child in root:
            if child.tag == "objective":
                objective = child
            elif child.tag == "reservation":
                reserved_value = float(child.attrib["value"])
            elif child.tag == "discount_factor":
                discount_factor = float(child.attrib["value"])

        if objective is None:
            if safe_parsing:
                pass
                # raise ValueError(f'No objective child was found in the root')
            objective = root
        weights = {}
        issues = {}
        real_issues = {}
        issue_info = {}
        issue_keys = {}
        rects, rect_utils = [], []

        def _get_hyperrects(ufun, max_utility, utiltype=float):
            utype = ufun.attrib.get("type", "none")
            uweight = float(ufun.attrib.get("weight", 1))
            uagg = ufun.attrib.get("aggregation", "sum")
            if uagg != "sum":
                raise ValueError(
                    f"Hypervolumes combined using {uagg} are not supported (only sum is supported)"
                )
            total_util = utiltype(0)
            rects = []
            rect_utils = []
            if utype == "PlainUfun":
                for rect in ufun:
                    util = utiltype(rect.attrib.get("utility", 0))
                    total_util += util if util > 0 else 0
                    ranges = {}
                    rect_utils.append(util * uweight)
                    for r in rect:
                        key = issue_keys[int(r.attrib["index"]) - 1]
                        ranges[key] = (
                            utiltype(r.attrib["min"]),
                            utiltype(r.attrib["max"]),
                        )
                    rects.append(ranges)
            else:
                raise ValueError(f"Unknown ufun type {utype}")
            total_util = total_util if not max_utility else max_utility
            if normalize_utility:
                for i, u in enumerate(rect_utils):
                    rect_utils[i] = u / total_util
            return rects, rect_utils

        for child in objective:
            if child.tag == "weight":
                indx = int(child.attrib["index"]) - 1
                weights[indx] = float(child.attrib["value"])
            elif child.tag == "utility_function" or child.tag == "utility":
                utility_tag = child
                max_utility = child.attrib.get("maxutility", None)
                if max_utility is not None:
                    max_utility = float(max_utility)
                ufun_found = False
                for ufun in utility_tag:
                    if ufun.tag == "ufun":
                        ufun_found = True
                        _r, _u = _get_hyperrects(ufun, max_utility)
                        rects += _r
                        rect_utils += _u
                if not ufun_found:
                    raise ValueError(
                        f"Cannot find ufun tag inside a utility_function tag"
                    )

            elif child.tag == "issue":
                indx = int(child.attrib["index"]) - 1
                myname = child.attrib["name"]
                issue_key = myname if keep_issue_names else indx
                if domain_issues is not None and myname not in domain_issues.keys():
                    raise ValueError(
                        f"Issue {myname} is not in the input issue names ({domain_issues.keys()})"
                    )
                issue_info[issue_key] = {"name": myname, "index": indx}
                issue_keys[indx] = issue_key
                info = {"type": "discrete", "etype": "discrete", "vtype": "discrete"}
                for a in ("type", "etype", "vtype"):
                    info[a] = child.attrib.get(a, info[a])
                mytype = info["type"]
                value_scale = None
                value_shift = None
                if mytype == "discrete":
                    issues[issue_key] = {}
                    if (
                        domain_issues is not None
                        and domain_issues[myname].is_uncountable()
                    ):
                        raise ValueError(
                            f"Got a {mytype} issue but expected a continuous valued issue"
                        )
                    # issues[indx]['items'] = {}
                elif mytype in ("integer", "real"):
                    lower, upper = (
                        child.attrib.get("lowerbound", None),
                        child.attrib.get("upperbound", None),
                    )
                    for rng_child in child:
                        if rng_child.tag == "range":
                            lower, upper = (
                                rng_child.attrib.get("lowerbound", lower),
                                rng_child.attrib.get("upperbound", upper),
                            )
                    if mytype == "integer":
                        issues[issue_key] = {}
                        if (
                            domain_issues is not None
                            and domain_issues[myname].is_uncountable()
                        ):
                            raise ValueError(
                                f"Got a {mytype} issue but expected a continuous valued issue"
                            )
                        # issues[indx]['items'] = {}
                        lower, upper = int(lower), int(upper)
                        for i in range(lower, upper + 1):
                            if domain_issues is not None and not outcome_is_valid(
                                (i,), [domain_issues[myname]]
                            ):
                                raise ValueError(
                                    f"Value {i} is not in the domain issue values: "
                                    f"{domain_issues[myname].values}"
                                )
                            issues[issue_key][i] = i if keep_value_names else i - lower
                    else:
                        lower, upper = float(lower), float(upper)
                        if (
                            domain_issues is not None
                            and not domain_issues[myname].is_uncountable()
                        ):
                            n_steps = domain_issues[myname].cardinality
                            delta = (n_steps - 1) / (upper - lower)
                            value_shift = -lower * delta
                            value_scale = delta
                            lower, upper = 0, n_steps - 1
                            issues[issue_key] = {}
                            for i in range(lower, upper + 1):
                                issues[issue_key][i] = (
                                    str(i) if keep_value_names else i - lower
                                )
                        else:
                            real_issues[issue_key] = {}
                            real_issues[issue_key]["range"] = (lower, upper)
                            real_issues[issue_key]["key"] = issue_key
                else:
                    raise ValueError(f"Unknown type: {mytype}")
                if mytype in "discrete" or "integer" or "real":
                    found_values = False
                    for item in child:
                        if item.tag == "item":
                            if mytype == "real":
                                raise ValueError(
                                    f"cannot specify item utilities for real type"
                                )
                            item_indx = int(item.attrib["index"]) - 1
                            item_name: str = item.attrib.get("value", None)
                            if item_name is None:
                                continue
                            item_key = (
                                item_name
                                if keep_value_names
                                and item_name is not None
                                and not force_numeric
                                else item_indx
                            )
                            if domain_issues is not None:
                                domain_all = list(domain_issues[myname].all)
                                if len(domain_all) > 0 and isinstance(
                                    domain_all[0], int
                                ):
                                    item_key = int(item_key)
                                if len(domain_all) > 0 and isinstance(
                                    domain_all[0], int
                                ):
                                    item_name = int(item_name)
                                if item_name not in domain_all:
                                    raise ValueError(
                                        f"Value {item_name} is not in the domain issue values: "
                                        f"{domain_issues[myname].values}"
                                    )
                                if len(domain_all) > 0 and isinstance(
                                    domain_all[0], int
                                ):
                                    item_name = str(item_name)
                            # TODO check that this casting is correct
                            if mytype == "integer":
                                item_key = int(item_key)
                            val = item.attrib.get("evaluation", None)
                            if val is None:
                                raise ValueError(
                                    f"Item {item_key} of issue {item_name} has not evaluation attribute!!"
                                )
                            issues[issue_key][item_key] = float(val)
                            found_values = True
                        elif item.tag == "evaluator":
                            if item.attrib["ftype"] == "linear":
                                offset = item.attrib.get(
                                    "offset", item.attrib.get("parameter0", 0.0)
                                )
                                slope = item.attrib.get(
                                    "slope", item.attrib.get("parameter1", 1.0)
                                )
                                offset, slope = float(offset), float(slope)
                                if value_scale is None:
                                    fun = lambda x: offset + slope * float(x)
                                else:
                                    fun = lambda x: offset + slope * (
                                        value_scale * float(x) + value_shift
                                    )
                            elif item.attrib["ftype"] == "triangular":
                                strt = item.attrib.get("parameter0", 0.0)
                                end = item.attrib.get("parameter1", 1.0)
                                middle = item.attrib.get("parameter2", 1.0)
                                strt, end, middle = (
                                    float(strt),
                                    float(end),
                                    float(middle),
                                )
                                offset1, slope1 = strt, (middle - strt)
                                offset2, slope2 = middle, (middle - end)
                                if value_scale is None:
                                    fun = (
                                        lambda x: offset1 + slope1 * float(x)
                                        if x < middle
                                        else offset2 + slope2 * float(x)
                                    )
                                else:
                                    fun = (
                                        lambda x: offset1
                                        + slope1
                                        * (value_scale * float(x) + value_shift)
                                        if x < middle
                                        else offset2
                                        + slope2
                                        * (value_scale * float(x) + value_shift)
                                    )
                            else:
                                raise ValueError(
                                    f'Unknown ftype {item.attrib["ftype"]}'
                                )
                            if mytype == "real" and value_scale is None:
                                real_issues[issue_key]["fun"] = fun
                            else:
                                for item_key, value in issues[issue_key].items():
                                    issues[issue_key][item_key] = fun(value)
                                found_values = True
                    if not found_values and issue_key in issues.keys():
                        issues.pop(issue_key, None)
                else:
                    """Here goes the code for real-valued issues"""

        if not keep_issue_names:
            issues = [issues[_] for _ in issues.keys()]
            real_issues = [real_issues[_] for _ in sorted(real_issues.keys())]
            for i, issue in enumerate(issues):
                issues[i] = [issue[_] for _ in issue.keys()]

        if safe_parsing and (
            len(weights) > 0
            and len(weights) != len(issues) + len(real_issues)
            and len(weights) != len(issues)
        ):
            raise ValueError(
                f"Got {len(weights)} weights for {len(issues)} issues and {len(real_issues)} real issues"
            )

        if force_single_issue and (
            len(rects) > 0
            or len(real_issues) > 1
            or (len(real_issues) > 0 and len(issues) > 0)
        ):
            raise ValueError(
                f"Cannot force single issue with a hyper-volumes based function"
            )

        # add utilities specified not as hyper-rectangles
        u = None
        if len(issues) > 0:
            if force_single_issue:
                if len(weights) > 0:
                    for key, issue in zip(ikeys(issues), ivalues(issues)):
                        try:
                            w = weights[issue_info[key]["index"]]
                        except:
                            w = 1.0
                        for item_key in ikeys(issue):
                            issue[item_key] *= w
                n_outcomes = None
                if max_n_outcomes is not None:
                    n_items = [len(_) for _ in ivalues(issues)]
                    n_outcomes = reduce(mul, n_items, 1)
                    if n_outcomes > max_n_outcomes:
                        return None, reserved_value, discount_factor
                if keep_value_names:
                    names = itertools.product(
                        *[
                            [
                                str(item_key).replace("&", "-")
                                for item_key in ikeys(items)
                            ]
                            for issue_key, items in zip(ikeys(issues), ivalues(issues))
                        ]
                    )
                    names = map(lambda items: ("+".join(items),), names)
                else:
                    if n_outcomes is None:
                        n_items = [len(_) for _ in ivalues(issues)]
                        n_outcomes = reduce(mul, n_items, 1)
                    names = [(_,) for _ in range(n_outcomes)]
                utils = itertools.product(
                    *[
                        [item_utility for item_utility in ivalues(items)]
                        for issue_key, items in zip(ikeys(issues), ivalues(issues))
                    ]
                )
                utils = map(lambda vals: sum(vals), utils)
                if normalize_utility:
                    utils = list(utils)
                    umax, umin = max(utils), (0.0 if normalize_max_only else min(utils))
                    if umax != umin:
                        utils = [(_ - umin) / (umax - umin) for _ in utils]
                if keep_issue_names:
                    u = MappingUtilityFunction(dict(zip(names, utils)))
                else:
                    u = MappingUtilityFunction(dict(zip(names, utils)))
            else:
                utils = None
                if normalize_utility:
                    utils = itertools.product(
                        *[
                            [
                                item_utility * weights[issue_info[issue_key]["index"]]
                                for item_utility in ivalues(items)
                            ]
                            for issue_key, items in zip(ikeys(issues), ivalues(issues))
                        ]
                    )
                    if len(weights) > 0:
                        ws = dict()
                        for key, issue in zip(ikeys(issues), ivalues(issues)):
                            try:
                                ws[key] = weights[issue_info[key]["index"]]
                            except:
                                ws[key] = 1.0
                        wsum = sum(weights.values())
                    else:
                        ws = [1.0] * len(issues)
                        wsum = len(issues)

                    utils = list(map(sum, utils))
                    umax, umin = max(utils), (0.0 if normalize_max_only else min(utils))
                    factor = umax - umin
                    if factor > 1e-8:
                        offset = umin / (wsum * factor)
                    else:
                        offset = 0.0
                        factor = umax if umax > 1e-8 else 1.0
                    for key, issue in ienumerate(issues):
                        for item_key in ikeys(issue):
                            issues[key][item_key] = (
                                issues[key][item_key] / factor - offset
                            )
                if len(issues) > 1:
                    ws = dict()
                    if len(weights) > 0:
                        for key, issue in zip(ikeys(issues), ivalues(issues)):
                            try:
                                ws[key] = weights[issue_info[key]["index"]]
                            except:
                                ws[key] = 1.0

                    if isinstance(issues, list):
                        ws = [ws[i] for i in range(len(issues))]

                    u = LinearUtilityAggregationFunction(
                        issue_utilities=issues, weights=ws
                    )
                else:
                    if len(weights) > 0:
                        for key, issue in zip(ikeys(issues), ivalues(issues)):
                            try:
                                w = weights[issue_info[key]["index"]]
                            except:
                                w = 1.0
                            for item_key in ikeys(issue):
                                issue[item_key] *= w
                    first_key = list(ikeys(issues))[0]
                    if utils is None:
                        utils = ivalues(issues[first_key])
                    if keep_issue_names:
                        u = MappingUtilityFunction(
                            dict(zip([(_,) for _ in ikeys(issues[first_key])], utils))
                        )
                    else:
                        u = MappingUtilityFunction(
                            dict(zip([(_,) for _ in range(len(utils))], utils))
                        )

        # add real_valued issues
        if len(real_issues) > 0:
            if len(weights) > 0:
                for key, issue in zip(ikeys(real_issues), ivalues(real_issues)):
                    try:
                        w = weights[issue_info[key]["index"]]
                    except:
                        w = 1.0
                    issue["fun_final"] = lambda x: w * issue["fun"](x)
            if normalize_utility:
                n_items_to_test = 10
                utils = itertools.product(
                    *[
                        [
                            issue["fun"](_)
                            for _ in np.linspace(
                                issue["range"][0],
                                issue["range"][1],
                                num=n_items_to_test,
                                endpoint=True,
                            )
                        ]
                        for key, issue in zip(ikeys(real_issues), ivalues(real_issues))
                    ]
                )
                if len(weights) > 0:
                    ws = dict()
                    for key, issue in zip(ikeys(issues), ivalues(issues)):
                        try:
                            ws[key] = weights[issue_info[key]["index"]]
                        except:
                            ws[key] = 1.0
                    wsum = sum(weights.values())
                else:
                    ws = [1.0] * len(issues)
                    wsum = len(issues)

                utils = list(map(lambda vals: sum(vals), utils))
                umax, umin = max(utils), (0.0 if normalize_max_only else min(utils))
                factor = umax - umin
                if factor > 1e-8:
                    offset = (umin / wsum) / factor
                else:
                    offset = 0.0
                    factor = 1.0
                for key, issue in real_issues.items():
                    issue["fun_final"] = lambda x: w * issue["fun"](x) / factor - offset
            u_real = LinearUtilityAggregationFunction(
                issue_utilities={_["key"]: _["fun_final"] for _ in real_issues.values()}
            )
            if u is None:
                u = u_real
            else:
                u = ComplexWeightedUtilityFunction(
                    ufuns=[u, u_real], weights=[1.0, 1.0]
                )

        # add hyper rectangles issues
        if len(rects) > 0:
            uhyper = HyperRectangleUtilityFunction(
                outcome_ranges=rects, utilities=rect_utils
            )
            if u is None:
                u = uhyper
            else:
                u = ComplexWeightedUtilityFunction(
                    ufuns=[u, uhyper], weights=[1.0, 1.0]
                )
        if not ignore_reserved and u is not None:
            u.reserved_value = reserved_value
        if ignore_discount:
            discount_factor = None
        return u, discount_factor

    def __getitem__(self, offer: "Outcome") -> Optional[UtilityValue]:
        """Overrides [] operator to call the ufun allowing it to act as a mapping"""
        return self(offer)

    def __call__(self, offer: "Outcome") -> UtilityValue:
        """Calculate the utility_function value for a given outcome.

        Args:
            offer: The offer to be evaluated.


        Remarks:

            - It calls the abstract method `eval` after opationally adjusting the
              outcome type.
            - It is preferred to override eval instead of directly overriding this method
            - You cannot return None from overriden eval() functions but raise an exception (ValueError) if it was
              not possible to calculate the UtilityValue.
            - Return A UtilityValue not a float for real-valued utilities for the benefit of inspection code.
            - Return the reserved value if the offer was None

        Returns:
            UtilityValue: The utility_function value which may be a distribution. If `None` it means the
                          utility_function value cannot be calculated.
        """
        if not self.issues and self.ami:
            self.issues = self.ami.issues
        if offer is None:
            return self.reserved_value
        if self.outcome_type is None:
            pass
        elif issubclass(self.outcome_type, tuple):
            offer = outcome_as_tuple(offer)
        elif issubclass(self.outcome_type, dict):
            offer = outcome_as_dict(offer, self.issue_names)
        else:
            if isinstance(offer, dict):
                offer = self.outcome_type(**offer)
            elif isinstance(offer, tuple):
                offer = self.outcome_type(*offer)
            else:
                offer = self.outcome_type(offer)
        return self.eval(offer)

    def eval(self, offer: "Outcome") -> UtilityValue:
        """Calculate the utility value for a given outcome.

        Args:
            offer: The offer to be evaluated.

        Returns:
            UtilityValue: The utility_function value which may be a distribution.
                          If `None` it means the utility_function value cannot
                          be calculated.

        Remarks:
            - You cannot return None from overriden eval() functions but
              raise an exception (ValueError) if it was
              not possible to calculate the UtilityValue.
            - Typehint the return type as a `UtilityValue` instead of a float
              for the benefit of inspection code.
            - Return the reserved value if the offer was None
            - *NEVER* call the baseclass using super() when overriding this
              method. Calling super will lead to an infinite loop.
            - The default implementation assumes that `is_better` is defined
              and uses it to do the evaluation. Note that the default
              implementation of `is_better` does assume that `eval` is defined
              and uses it. This means that failing to define both leads to
              an infinite loop.
        """
        raise NotImplementedError("Could not calculate the utility value.")

    def eval_all(self, outcomes: List["Outcome"]) -> Iterable[UtilityValue]:
        """
        Calculates the utility value of a list of outcomes and returns their
        utility values

        Args:
            outcomes: A list of offers

        Returns:
            An iterable with the utility values of the given outcomes in order

        Remarks:
            - The default implementation just iterates over the outcomes
              calling the ufun for each of them. In a distributed environment,
              it is possible to do this in parallel using a thread-pool for example.
        """
        return [self(_) for _ in outcomes]

    @classmethod
    def approximate(
        cls,
        ufuns: List["UtilityFunction"],
        issues: Iterable["Issue"],
        n_outcomes: int,
        min_per_dim=5,
        force_single_issue=False,
    ) -> Tuple[List["MappingUtilityFunction"], List["Outcome"], List["Issue"]]:
        """
        Approximates a list of ufuns with a list of mapping discrete ufuns

        Args:
            ufuns: The list of ufuns to approximate
            issues: The issues
            n_outcomes: The number of outcomes to use in the approximation
            min_per_dim: Minimum number of levels per continuous dimension
            force_single_issue: Force the output to have a single issue

        Returns:

        """
        issues = list(issues)
        issue_names = [_.name for _ in issues]
        outcomes = sample_outcomes(
            issues=issues,
            astype=tuple,
            min_per_dim=min_per_dim,
            expansion_policy="null",
            n_outcomes=n_outcomes,
        )
        if force_single_issue:
            output_outcomes = [(_,) for _ in range(n_outcomes)]
            output_issues = [Issue(values=len(output_outcomes))]
        else:
            output_outcomes = outcomes
            issue_values = []
            for i in range(len(issues)):
                vals = np.unique(np.array([_[i] for _ in outcomes])).tolist()
                issue_values.append(vals)
            output_issues = [
                Issue(name=issue.name, values=issue_vals)
                for issue, issue_vals in zip(issues, issue_values)
            ]

        utils = []
        for ufun in ufuns:
            u = [ufun(outcome_as(o, ufun.outcome_type, issue_names)) for o in outcomes]
            utils.append(MappingUtilityFunction(mapping=dict(zip(output_outcomes, u))))

        return utils, output_outcomes, output_issues

    def compare_real(self, o1: "Outcome", o2: "Outcome", method="mean") -> float:
        """
        Compares the two outcomes and returns a measure of the difference
        between their utilities.

        Args:
            o1: First outcome
            o2: Second outcome
            method: The comparison method if one of the two outcomes result in
                    a distribution. Acceptable values are:

                        - mean: Compares the means of the two distributions
                        - min: Compares minimum values with nonzero probability
                        - max: Compares maximum values with nonzero probability
                        - int: Calculates :math:`int (u_1-u_2) du_1du_2`.
                        - Callable: The callable is given u(o1), u(o2) and
                                    should return the comparison.

        """
        u1, u2 = self(o1), self(o2)
        if isinstance(u1, float) and isinstance(u2, float):
            return u1 - u2
        if isinstance(method, Callable):
            return method(u1, u2)
        if isinstance(u1, float):
            u1 = UtilityDistribution(dtype="uniform", loc=u1, scale=1e-10)
        if isinstance(u2, float):
            u2 = UtilityDistribution(dtype="uniform", loc=u2, scale=1e-10)
        if method == "mean":
            return u1.mean() - u2.mean()
        if method == "min":
            return u1.min() - u2.min()
        if method == "max":
            return u1.max() - u2.max()
        if method == "int":
            if u1.scale <= 1e-9:
                return u1 - u2.mean()
            if u2.scale <= 1e-9:
                return u1.mean() - u2
        raise NotImplementedError(
            "Should calculate the integration [(u1-u2) du1 du2] but not implemented yet."
        )

    def rank_with_weights(
        self, outcomes: List[Optional[Outcome]], descending=True
    ) -> List[Tuple[int, float]]:
        """Ranks the given list of outcomes with weights. None stands for the null outcome.

        Returns:

            - A list of tuples each with two values:
                - an integer giving the index in the input array (outcomes) of an outcome
                - the weight of that outcome
            - The list is sorted by weights descendingly

        """
        return sorted(
            zip(list(range(len(outcomes))), [float(self(o)) for o in outcomes]),
            key=lambda x: x[1],
            reverse=descending,
        )

    def argsort(self, outcomes: List[Optional[Outcome]], descending=True) -> List[int]:
        """Finds the rank of each outcome as an integer"""
        return [_[0] for _ in self.rank_with_weights(outcomes, descending=descending)]

    def sort(
        self, outcomes: List[Optional[Outcome]], descending=True
    ) -> List[Optional[Outcome]]:
        """Sorts the given outcomes in place in ascending or descending order of utility value.

        Returns:
            Returns the input list after being sorted. Notice that the array is sorted in-place

        """
        outcomes.sort(key=self, reverse=descending)
        return outcomes

    rank = argsort
    """Ranks the given list of outcomes. None stands for the null outcome"""

    def is_better(
        self, first: "Outcome", second: "Outcome", epsilon=1e-10
    ) -> Optional[bool]:
        """
        Compares two offers using the `ufun` returning whether the first is better than the second

        Args:
            first: First outcome to be compared
            second: Second outcome to be compared
            epsilon: comparison threshold. If the utility difference within the range [-epsilon, epsilon] the two
                     outcomes are assumed to be compatible

        Returns:
            True if utility(first) > utility(second) + epsilon
            None if |utility(first) - utility(second)| <= epsilon
            False if utility(first) < utility(second) - epsilon
        """
        u1, u2 = self(first), self(second)
        if u1 is None or u2 is None or abs(u1 - u2) <= epsilon:
            return None
        return float(u1) > float(u2)

    @abstractmethod
    def xml(self, issues: List[Issue]) -> str:
        """Converts the function into a well formed XML string preferrably in GENIUS format.

        If the output has with </objective> then discount factor and reserved value should also be included
        If the output has </utility_space> it will not be appended in `to_xml_str`

        """

    @property
    def type(self) -> str:
        """Returns the utility_function type.

        Each class inheriting from this ``UtilityFunction`` class will have its own type. The default type is the empty
        string.

        Examples:
            >>> from negmas.utilities import *
            >>> print(LinearUtilityAggregationFunction({1:lambda x:x, 2:lambda x:x}).type)
            linear_aggregation
            >>> print(MappingUtilityFunction(lambda x: x).type)
            mapping
            >>> print(NonLinearUtilityAggregationFunction({1:lambda x:x}, f=lambda x: x).type)
            non_linear_aggregation

        Returns:
            str: utility_function type
        """
        return snake_case(
            self.__class__.__name__.replace("Function", "").replace("Utility", "")
        )

    @property
    def base_type(self) -> str:
        """Returns the utility_function base type ignoring discounting and similar wrappings."""
        return self.type

    def eu(self, offer: "Outcome") -> Optional[float]:
        """Calculate the expected utility value.

        Args:
            offer: The offer to be evaluated.

        Returns:
            float: The expected utility value for UFuns that return a distribution and just utility value for real-valued utilities.

        """
        v = self(offer)
        return float(v) if v is not None else None

    @classmethod
    def generate_bilateral(
        cls,
        outcomes: Union[int, List[Outcome]],
        conflict_level: float = 0.5,
        conflict_delta=0.005,
        win_win=0.5,
    ) -> Tuple["UtilityFunction", "UtilityFunction"]:
        """Generates a couple of utility functions

        Args:

            n_outcomes (int): number of outcomes to use
            conflict_level: How conflicting are the two ufuns to generate.
                            1.0 means maximum conflict.
            conflict_delta: How variable is the conflict at different outcomes.
            win_win: How much are their opportunities for win-win situations.

        Examples:

            >>> u1, u2 = UtilityFunction.generate_bilateral(outcomes=10, conflict_level=0.0
            ...                                             , conflict_delta=0.0, win_win=0.0)
            >>> print(UtilityFunction.conflict_level(u1, u2, outcomes=10))
            0.0

            >>> u1, u2 = UtilityFunction.generate_bilateral(outcomes=10, conflict_level=1.0
            ...                                             , conflict_delta=0.0, win_win=0.0)
            >>> print(UtilityFunction.conflict_level(u1, u2, outcomes=10))
            1.0

            >>> u1, u2 = UtilityFunction.generate_bilateral(outcomes=10, conflict_level=0.5
            ...                                             , conflict_delta=0.0, win_win=1.0)
            >>> 0.0 <= UtilityFunction.conflict_level(u1, u2, outcomes=10) <= 1.0
            True


        """
        from negmas.utilities.nonlinear import MappingUtilityFunction

        if isinstance(outcomes, int):
            outcomes = [(_,) for _ in range(outcomes)]
        n_outcomes = len(outcomes)
        u1 = np.random.random(n_outcomes)
        rand = np.random.random(n_outcomes)
        if conflict_level > 0.5:
            conflicting = 1.0 - u1 + conflict_delta * np.random.random(n_outcomes)
            u2 = conflicting * conflict_level + rand * (1 - conflict_level)
        elif conflict_level < 0.5:
            same = u1 + conflict_delta * np.random.random(n_outcomes)
            u2 = same * (1 - conflict_level) + rand * conflict_level
        else:
            u2 = rand

        # todo implement win_win correctly. Order the ufun then make outcomes with good outcome even better and vice
        # versa
        # u2 += u2 * win_win
        # u2 += np.random.random(n_outcomes) * conflict_delta
        u1 -= u1.min()
        u2 -= u2.min()
        u1 = u1 / u1.max()
        u2 = u2 / u2.max()
        if random.random() > 0.5:
            u1, u2 = u2, u1
        return (
            MappingUtilityFunction(dict(zip(outcomes, u1))),
            MappingUtilityFunction(dict(zip(outcomes, u2))),
        )

    @classmethod
    def generate_random_bilateral(
        cls, outcomes: Union[int, List[Outcome]]
    ) -> Tuple["UtilityFunction", "UtilityFunction"]:
        """Generates a couple of utility functions

        Args:

            n_outcomes (int): number of outcomes to use
            conflict_level: How conflicting are the two ufuns to generate. 1.0 means maximum conflict.
            conflict_delta: How variable is the conflict at different outcomes.
            zero_summness: How zero-sum like are the two ufuns.


        """
        from negmas.utilities.nonlinear import MappingUtilityFunction

        if isinstance(outcomes, int):
            outcomes = [(_,) for _ in range(outcomes)]
        n_outcomes = len(outcomes)
        u1 = np.random.random(n_outcomes)
        u2 = np.random.random(n_outcomes)
        u1 -= u1.min()
        u2 -= u2.min()
        u1 /= u1.max()
        u2 /= u2.max()
        return (
            MappingUtilityFunction(
                dict(zip(outcomes, u1)), outcome_type=type(outcomes[0])
            ),
            MappingUtilityFunction(
                dict(zip(outcomes, u2)), outcome_type=type(outcomes[0])
            ),
        )

    @classmethod
    def generate_random(
        cls, n: int, outcomes: Union[int, List[Outcome]], normalized: bool = True
    ) -> List["UtilityFunction"]:
        """Generates N mapping utility functions

        Args:
            n: number of utility functions to generate
            outcomes: number of outcomes to use
            normalized: if true, the resulting ufuns will be normlized between zero and one.


        """
        from negmas.utilities.nonlinear import MappingUtilityFunction

        if isinstance(outcomes, int):
            outcomes = [(_,) for _ in range(outcomes)]
        n_outcomes = len(outcomes)
        ufuns = []
        for _ in range(n):
            u1 = np.random.random(n_outcomes)
            if normalized:
                u1 -= u1.min()
                u1 /= u1.max()
            ufuns.append(MappingUtilityFunction(dict(zip(outcomes, u1))))
        return ufuns

    @classmethod
    def opposition_level(
        cls,
        ufuns=List["UtilityFunction"],
        max_utils: Union[float, Tuple[float, float]] = 1.0,
        outcomes: Union[int, List["Outcome"]] = None,
        issues: List["Issue"] = None,
        max_tests: int = 10000,
    ) -> float:
        """
        Finds the opposition level of the two ufuns defined as the minimum distance to outcome (1, 1)

        Args:
            ufuns: A list of utility functions to use.
            max_utils: A list of maximum utility value for each ufun (or a single number if they are equal).
            outcomes: A list of outcomes (should be the complete issue space) or an integer giving the number
                     of outcomes. In the later case, ufuns should expect a tuple of a single integer.
            issues: The issues (only used if outcomes is None).
            max_tests: The maximum number of outcomes to use. Only used if issues is given and has more
                       outcomes than this value.


        Examples:


            - Opposition level of the same ufun repeated is always 0
            >>> from negmas.utilities.nonlinear import MappingUtilityFunction
            >>> u1, u2 = lambda x: x[0], lambda x: x[0]
            >>> UtilityFunction.opposition_level([u1, u2], outcomes=10, max_utils=9)
            0.0

            - Opposition level of two ufuns that are zero-sum
            >>> u1, u2 = MappingUtilityFunction(lambda x: x[0]), MappingUtilityFunction(lambda x: 9 - x[0])
            >>> UtilityFunction.opposition_level([u1, u2], outcomes=10, max_utils=9)
            0.7114582486036499

        """
        if outcomes is None and issues is None:
            raise ValueError("You must either give outcomes or issues")
        if outcomes is None:
            outcomes = Issue.enumerate(issues, max_n_outcomes=max_tests, astype=tuple)
        if isinstance(outcomes, int):
            outcomes = [(_,) for _ in range(outcomes)]
        if not isinstance(max_utils, Iterable):
            max_utils = [max_utils] * len(ufuns)
        if len(ufuns) != len(max_utils):
            raise ValueError(
                f"Cannot use {len(ufuns)} ufuns with only {len(max_utils)} max. utility values"
            )

        nearest_val = float("inf")
        assert not any(abs(_) < 1e-7 for _ in max_utils), f"max-utils : {max_utils}"
        for outcome in outcomes:
            v = sum(
                (1.0 - u(outcome) / max_util) ** 2
                for max_util, u in zip(max_utils, ufuns)
            )
            if v == float("inf"):
                warnings.warn(
                    f"u is infinity: {outcome}, {[_(outcome) for _ in ufuns]}, max_utils"
                )
            if v < nearest_val:
                nearest_val = v
        return sqrt(nearest_val)

    @classmethod
    def conflict_level(
        cls,
        u1: "UtilityFunction",
        u2: "UtilityFunction",
        outcomes: Union[int, List["Outcome"]],
        max_tests: int = 10000,
    ) -> float:
        """
        Finds the conflict level in these two ufuns

        Args:
            u1: first utility function
            u2: second utility function

        Examples:
            - A nonlinear strictly zero sum case
            >>> from negmas.utilities.nonlinear import MappingUtilityFunction
            >>> outcomes = [(_,) for _ in range(10)]
            >>> u1 = MappingUtilityFunction(dict(zip(outcomes,
            ... np.random.random(len(outcomes)))))
            >>> u2 = MappingUtilityFunction(dict(zip(outcomes,
            ... 1.0 - np.array(list(u1.mapping.values())))))
            >>> print(UtilityFunction.conflict_level(u1=u1, u2=u2, outcomes=outcomes))
            1.0

            - The same ufun
            >>> print(UtilityFunction.conflict_level(u1=u1, u2=u1, outcomes=outcomes))
            0.0

            - A linear strictly zero sum case
            >>> outcomes = [(i,) for i in range(10)]
            >>> u1 = MappingUtilityFunction(dict(zip(outcomes,
            ... np.linspace(0.0, 1.0, len(outcomes), endpoint=True))))
            >>> u2 = MappingUtilityFunction(dict(zip(outcomes,
            ... np.linspace(1.0, 0.0, len(outcomes), endpoint=True))))
            >>> print(UtilityFunction.conflict_level(u1=u1, u2=u2, outcomes=outcomes))
            1.0
        """
        if isinstance(outcomes, int):
            outcomes = [(_,) for _ in range(outcomes)]
        n_outcomes = len(outcomes)
        points = np.array([[u1(o), u2(o)] for o in outcomes])
        order = np.random.permutation(np.array(range(n_outcomes)))
        p1, p2 = points[order, 0], points[order, 1]
        signs = []
        trial = 0
        for i in range(n_outcomes - 1):
            for j in range(i + 1, n_outcomes):
                if trial >= max_tests:
                    break
                trial += 1
                o11, o12 = p1[i], p1[j]
                o21, o22 = p2[i], p2[j]
                if o12 == o11 and o21 == o22:
                    continue
                signs.append(
                    int((o12 > o11 and o21 > o22) or (o12 < o11 and o21 < o22))
                )
        signs = np.array(signs)
        if len(signs) == 0:
            return None
        return signs.mean()

    @classmethod
    def winwin_level(
        cls,
        u1: "UtilityFunction",
        u2: "UtilityFunction",
        outcomes: Union[int, List["Outcome"]],
        max_tests: int = 10000,
    ) -> float:
        """
        Finds the win-win level in these two ufuns

        Args:
            u1: first utility function
            u2: second utility function

        Examples:
            - A nonlinear same ufun case
            >>> from negmas.utilities.nonlinear import MappingUtilityFunction
            >>> outcomes = [(_,) for _ in range(10)]
            >>> u1 = MappingUtilityFunction(dict(zip(outcomes,
            ... np.linspace(1.0, 0.0, len(outcomes), endpoint=True))))

            - A linear strictly zero sum case
            >>> outcomes = [(_,) for _ in range(10)]
            >>> u1 = MappingUtilityFunction(dict(zip(outcomes,
            ... np.linspace(0.0, 1.0, len(outcomes), endpoint=True))))
            >>> u2 = MappingUtilityFunction(dict(zip(outcomes,
            ... np.linspace(1.0, 0.0, len(outcomes), endpoint=True))))


        """
        if isinstance(outcomes, int):
            outcomes = [(_,) for _ in range(outcomes)]
        n_outcomes = len(outcomes)
        points = np.array([[u1(o), u2(o)] for o in outcomes])
        order = np.random.permutation(np.array(range(n_outcomes)))
        p1, p2 = points[order, 0], points[order, 1]
        signed_diffs = []
        for trial, (i, j) in enumerate(
            zip(range(n_outcomes - 1), range(1, n_outcomes))
        ):
            if trial >= max_tests:
                break
            o11, o12 = p1[i], p1[j]
            o21, o22 = p2[i], p2[j]
            if o11 == o12:
                if o21 == o22:
                    continue
                else:
                    win = abs(o22 - o21)
            elif o11 < o12:
                if o21 == o22:
                    win = o12 - o11
                else:
                    win = (o12 - o11) + (o22 - o21)
            else:
                if o21 == o22:
                    win = o11 - o12
                else:
                    win = (o11 - o12) + (o22 - o21)
            signed_diffs.append(win)
        signed_diffs = np.array(signed_diffs)
        if len(signed_diffs) == 0:
            return None
        return signed_diffs.mean()

    def outcome_with_utility(
        self,
        rng: Tuple[Optional[float], Optional[float]],
        issues: List[Issue] = None,
        outcomes: List[Outcome] = None,
        n_trials: int = 100,
    ) -> Optional["Outcome"]:
        """
        Gets one outcome within the given utility range or None on failure

        Args:
            self: The utility function
            rng: The utility range
            issues: The issues the utility function is defined on
            outcomes: The outcomes to sample from
            n_trials: The maximum number of trials

        Returns:

            - Either issues, or outcomes should be given but not both

        """
        if outcomes is None:
            outcomes = Issue.sample(
                issues=issues,
                n_outcomes=n_trials,
                astype=self.outcome_type,
                with_replacement=False,
                fail_if_not_enough=False,
            )
        n = min(len(outcomes), n_trials)
        mn, mx = rng
        if mn is None:
            mn = float("-inf")
        if mx is None:
            mx = float("inf")
        for i in range(n):
            o = outcomes[i]
            if mn <= self(o) <= mx:
                return o
        return None

    def utility_range(
        self,
        issues: List[Issue] = None,
        outcomes: Collection[Outcome] = None,
        infeasible_cutoff: Optional[float] = None,
        return_outcomes=False,
        max_n_outcomes=1000,
        ami: Optional["AgentMechnismInterface"] = None,
    ) -> Union[
        Tuple[UtilityValue, UtilityValue],
        Tuple[UtilityValue, UtilityValue, Outcome, Outcome],
    ]:
        """Finds the range of the given utility function for the given outcomes

        Args:
            self: The utility function
            issues: List of issues (optional)
            outcomes: A collection of outcomes (optional)
            infeasible_cutoff: A value under which any utility is considered infeasible and is not used in calculation
            return_outcomes: If true, will also return an outcome for min and max utils
            max_n_outcomes: the maximum number of outcomes to try sampling (if sampling is used and outcomes are not
                            given)
            ami: Optional AMI to use

        Returns:
            UtilityFunction: A utility function that is guaranteed to be normalized for the set of given outcomes

        """

        if outcomes is None:
            outcomes = Issue.sample(
                issues,
                n_outcomes=max_n_outcomes,
                with_replacement=True,
                fail_if_not_enough=False,
                astype=self.outcome_type,
            )
        outcomes = [_ for _ in outcomes if _ is not None]
        utils = [self(o) for o in outcomes]
        errors = [i for i, u in enumerate(utils) if u is None]
        if len(errors) > 0:
            raise ValueError(
                f"UFun returnd None for {len(errors) / len(utils):03%} outcomes\n"
                # f"outcomes {[outcomes[e] for e in errors]}\n"
            )
        # if there are no outcomes return zeros for utils
        if len(utils) == 0:
            if return_outcomes:
                return 0.0, 0.0, None, None
            return 0.0, 0.0

        # make sure the utility value is converted to float
        utils = [float(_) for _ in utils]

        # if there is an infeasible_cutoff, apply it
        if infeasible_cutoff is not None:
            if return_outcomes:
                outcomes = [o for o, _ in zip(outcomes, utils) if _ > infeasible_cutoff]
            utils = np.array([_ for _ in utils if _ > infeasible_cutoff])

        if return_outcomes:
            minloc, maxloc = np.argmin(utils), np.argmax(utils)
            return (
                utils[minloc],
                utils[maxloc],
                outcomes[minloc],
                outcomes[maxloc],
            )
        return float(np.min(utils)), float(np.max(utils))

    def init_inverse(
        self,
        issues: Optional[List["Issue"]] = None,
        outcomes: Optional[List["Outcome"]] = None,
        n_trials: int = 10000,
        max_cache_size: int = 10000,
    ) -> None:
        """
        Initializes the inverse ufun used to map utilities to outcomes.

        Args:
            issues: The issue space to be searched for inverse. If not given, the issue space of the mechanism may be used.
            outcomes: The outcomes to consider. This takes precedence over `issues`
            n_trials: Used for constraining computation if necessary
            max_cache_size: The maximum allowed number of outcomes to cache

        """
        if issues is None:
            issues = self.issues
        self._min, self._max, self._worst, self._best = self.utility_range(
            issues, outcomes, return_outcomes=True, max_n_outcomes=n_trials
        )
        self._range = self._max - self._min
        self._offset = self._min / self._range if self._range > 1e-5 else self._min
        astype = self._outcome_type if self._outcome_type else tuple
        if self._issues and Issue.num_outcomes(issues) < max_cache_size:
            outcomes = Issue.enumerate(issues, astype=astype)
        if not outcomes:
            outcomes = Issue.discretize_and_enumerate(
                issues, n_discretization=2, astype=astype, max_n_outcomes=max_cache_size
            )
            n = max_cache_size - len(outcomes)
            if n > 0:
                outcomes += Issue.sample(
                    issues,
                    n,
                    astype=astype,
                    with_replacement=False,
                    fail_if_not_enough=False,
                )
        utils = self.eval_all(outcomes)
        self._ordered_outcomes = sorted(zip(utils, outcomes), key=lambda x: -x[0])
        self._inverse_initialzed = True

    def inverse(
        self,
        u: float,
        eps: Union[float, Tuple[float, float]] = (1e-3, 0.2),
        assume_normalized=True,
        issues: Optional[List["Issue"]] = None,
        outcomes: Optional[List["Outcome"]] = None,
        n_trials: int = 10000,
        return_all_in_range=False,
        max_n_outcomes=10000,
    ) -> Union[Optional["Outcome"], Tuple[List["Outcome"], List[UtilityValue]]]:
        """
        Finds an outcmoe with the given utility value

        Args:
            u: the utility value to find an outcome for.
            eps: An approximation error. If a number, it is mapped to (eps, eps)
            n_trials: the number of time to try (used if random sampling is employed)
            issues: If given the issue space to search (not recommended)
            outcomes: If given the outcomes to search (not recommended)
            assume_normalized: If true, the ufun will be assumed normalized between 0 and 1 (faster)
            return_all_in_range: If given all outcomes in the given range (or samples of them) will be returned up
                                 to `max_n_outcomes`
            max_n_outcomes: Only used if return_all_in_range is given and gives the maximum number of outcomes to return

        Returns:
            - if return_all_in_range:
               - A tuple with all outcomes in the given range (or samples thereof)
               - A tuple of corresponding utility values
            - Otherwise, An outcome with utility between u-eps[0] and u+eps[1] if found

        Remarks:
            - If issues or outcomes are not None, then init_inverse will be called first

        """
        if not isinstance(eps, Iterable):
            mn, mx = u - eps, u + eps
        else:
            mn, mx = u - eps[0], u + eps[1]
        if not assume_normalized:
            if self._range > 1e-5:
                mn = mn / self._range + self._offset
                mx = mx / self._range + self._offset
            else:
                mn = mn + self._offset
                mx = mx + self._offset
        if (not self._inverse_initialzed) or issues or outcomes:
            self.init_inverse(issues, outcomes, n_trials, n_trials)
        # todo use bisection
        if return_all_in_range:
            results = []
            utils = []
            for i, (util, w) in enumerate(self._ordered_outcomes):
                if utils > mx:
                    continue
                if util < mn or len(results) >= max_n_outcomes:
                    break
                results.append(w)
                utils.append(util)
            return results, utils

        for i, (util, w) in enumerate(self._ordered_outcomes):
            if util >= mn:
                continue
            ubefore, wbefore = self._ordered_outcomes[i - 1 if i > 0 else 0]
            if ubefore > mx:
                return None
            return wbefore
        ubefore, wbefore = self._ordered_outcomes[-1]
        if ubefore > mx:
            return None
        return wbefore

    @property
    def is_inverse_initialized(self):
        return self._inverse_initialzed

    def uninialize_inverse(self):
        self._inverse_initialzed = False

    @classmethod
    def random(
        cls, issues, reserved_value=(0.0, 0.5), normalized=True, **kwargs
    ) -> "UtilityFunction":
        """Generates a random ufun of the given type"""

    @classmethod
    def from_str(cls, s) -> "UtilityFunction":
        """Creates an object out of a dict."""
        return deserialize(eval(s))

    def __str__(self):
        return str(serialize(self))


UtilityFunctions = List["UtilityFunction"]



"""
Implements serialization to and from strings and secondary storage.

"""
import warnings
import json
from typing import Any, Dict, Iterable, Optional
import cloudpickle

import numpy as np
from pandas import json_normalize

from .helpers import (
    get_class,
    is_lambda_function,
    is_non_lambda_function,
    is_jsonable,
    dump,
    load,
)

__all__ = [
    "serialize",
    "deserialize",
    "dump",
    "load",
    "to_flat_dict",
    "PYTHON_CLASS_IDENTIFIER",
]

PYTHON_CLASS_IDENTIFIER = "__python_class__"
LAMBDA_START = b"__LAMBDAOBJ__:"
FUNCTION_START = b"__FUNCTION_START__:"
# JSON_START = b"__JSON_START__:"
CLOUDPICKLE_START = b"__CLOUDPICKLE_START__:"
SPECIAL_FIELDS = ("_NamedObject__uuid", "_NamedObject__name")
SPECIAL_FIELDS_SHORT_NAMES = ("id", "name")


def to_flat_dict(value, deep=True, add_type_field=False) -> Dict[str, Any]:
    """
    Encodes the given value as a flat dictionary

    Args:
        value: The value to be converted to a flat dictionary
        deep: Converting all sub-objects
        add_type_field: If true, a special field for the object type will be added

    Returns:

    """
    d = serialize(value, add_type_field=add_type_field, deep=deep)
    if d is None:
        return {}
    if not isinstance(d, dict):
        raise ValueError(
            f"value is of type {type(value)} cannot be converted to a flat dict"
        )
    for k, v in d.items():
        if isinstance(v, list) or isinstance(v, tuple):
            d[k] = str(v)
    return json_normalize(d, errors="ignore", sep="_").to_dict(orient="records")[0]


def serialize(
    value,
    deep=True,
    add_type_field=True,
    keep_private=False,
    ignore_methods=True,
    ignore_lambda=False,
    objmem=None,
):
    """Encodes the given value as nothing more complex than simple dict
    of either dicts, lists or builtin numeric or string values. The resulting
    dictionary will be json serializable

    Args:

        value: Any object
        deep: Whether we should go deep in the encoding or do a shallow encoding
        add_type_field: Whether to add a type field. If True, A field named `PYTHON_CLASS_IDENTIFIER` will be added
        giving the type of `value`
        keep_private: Keeps fields starting with "_"

    Remarks:

        - All iterables are converted to lists when `deep` is true.
        - If the `value` object has a `to_dict` member, it will be called to
          do the conversion, otherwise its `__dict__` or `__slots__` member
          will be used.

    See Also:
          `deserialize`, `PYTHON_CLASS_IDENTIFIER`

    """

    def add_to_mem(x, objmem):
        if not objmem:
            objmem = {id(x)}
        else:
            objmem.add(id(x))
        return objmem

    def good_field(k: str, v, objmem):
        if objmem and id(v) in objmem:
            return False
        if ignore_methods and is_non_lambda_function(v):
            return False
        if ignore_lambda and is_lambda_function(v):
            return False
        if not isinstance(k, str):
            return False
        return keep_private or not (k != PYTHON_CLASS_IDENTIFIER and k.startswith("_"))

    def adjust_dict(d):
        if not isinstance(d, dict):
            return d
        for a, b in zip(SPECIAL_FIELDS, SPECIAL_FIELDS_SHORT_NAMES):
            if a in d.keys():
                if b in d.keys() and d[b] != d[a]:
                    warnings.warn(f"Field {a} and {b} already exist and are not equal.")
                d[b] = d[a]
                del d[b]
        return d

    if value is None:
        return None

    if isinstance(value, dict):
        if not deep:
            return adjust_dict({k: v for k, v in value.items()})
        return adjust_dict(
            {
                k: serialize(v, deep=deep, add_type_field=add_type_field, objmem=objmem)
                for k, v in value.items()
                if good_field(k, v, objmem)
            }
        )
    if isinstance(value, Iterable) and not deep:
        # add_to_mem(value)
        return value
    if isinstance(value, Iterable) and not isinstance(value, str):
        objmem = add_to_mem(value, objmem)
        return adjust_dict(
            type(value)(
                (
                    serialize(
                        _, deep=deep, add_type_field=add_type_field, objmem=objmem
                    )
                    for _ in value
                )
            )
        )
    if hasattr(value, "to_dict"):
        converted = value.to_dict()
        if isinstance(converted, dict):
            if add_type_field and (PYTHON_CLASS_IDENTIFIER not in converted.keys()):
                converted[PYTHON_CLASS_IDENTIFIER] = (
                    value.__class__.__module__ + "." + value.__class__.__name__
                )
            return adjust_dict({k: v for k, v in converted.items()})
        else:
            return adjust_dict(converted)
    if isinstance(value, str):
        return value
    if isinstance(value, bytes):
        if (
            value.startswith(FUNCTION_START)
            or value.startswith(LAMBDA_START)
            or value.startswith(CLOUDPICKLE_START)
            # or value.startswith(JSON_START)
        ):
            warnings.warn(
                f"{value} starts with a reserved part!! Will just keep it as"
                f" it is. May be you are serializing an already serialized object"
            )
        return value

    if is_lambda_function(value):
        return LAMBDA_START + cloudpickle.dumps(value)

    if is_non_lambda_function(value):
        return FUNCTION_START + cloudpickle.dumps(value)

    if hasattr(value, "__dict__"):
        if deep:
            objmem = add_to_mem(value, objmem)
            d = {
                k: serialize(v, deep=deep, add_type_field=add_type_field, objmem=objmem)
                for k, v in value.__dict__.items()
                if good_field(k, v, objmem)
            }
        else:
            d = {k: v for k, v in value.__dict__.items() if good_field(k, v, objmem)}
        if add_type_field:
            d[PYTHON_CLASS_IDENTIFIER] = (
                value.__class__.__module__ + "." + value.__class__.__name__
            )
        return adjust_dict(d)

    if hasattr(value, "__slots__"):
        if deep:
            objmem = add_to_mem(value, objmem)
            d = dict(
                zip(
                    (k for k in value.__slots__),
                    (
                        serialize(
                            getattr(value, _),
                            deep=deep,
                            add_type_field=add_type_field,
                            objmem=objmem,
                        )
                        for _ in value.__slots__
                    ),
                )
            )
        else:
            d = dict(
                zip(
                    (k for k in value.__slots__),
                    (getattr(value, _) for _ in value.__slots__),
                )
            )
        if add_type_field:
            d[PYTHON_CLASS_IDENTIFIER] = (
                value.__class__.__module__ + "." + value.__class__.__name__
            )
        return adjust_dict(d)
    if isinstance(value, np.int64):
        return int(value)
    # a builtin
    if is_jsonable(value):
        return value
    try:
        vv = CLOUDPICKLE_START + cloudpickle.dumps(value)
        return vv
    except:
        pass
    warnings.warn(f"{value} of type {type(value)} is not serializable")
    return value


def deserialize(
    d: Any,
    deep=True,
    remove_type_field=True,
    keep_private=False,
    fallback_class_name: Optional[str] = None,
):
    """Decodes a dict/object coming from `serialize`

    Args:

        d: The value to be decoded. If it is not a dict, it is returned as it is.
        deep: If true, decode recursively
        remove_type_field: If true the field called `PYTHON_CLASS_IDENTIFIER` will be removed if found.
        keep_private: If given, private fields (starting with _) will be kept
        fallback_class_name: If given, it is used as the fall-back  type if ``PYTHON_CLASS_IDENTIFIER` is not in the dict.

    Remarks:

        - If the object is not a dict or if it has no `PYTHON_CLASS_IDENTIFIER` field and no `fallback_class_name` is
          given, the input `d` is returned as it is. It will not even be copied.

    See Also:
        `serialize`, `PYTHON_CLASS_IDENTIFIER`



    """

    def good_field(k: str):
        return keep_private or not (k != PYTHON_CLASS_IDENTIFIER and k.startswith("_"))

    if d is None or isinstance(d, int) or isinstance(d, float) or isinstance(d, str):
        return d
    if isinstance(d, dict):
        if remove_type_field:
            python_class_name = d.pop(PYTHON_CLASS_IDENTIFIER, fallback_class_name)
        else:
            python_class_name = d.get(PYTHON_CLASS_IDENTIFIER, fallback_class_name)
        if python_class_name is not None:
            python_class_name = python_class_name
            if python_class_name.endswith("Issue"):
                python_class = get_class("negmas.outcomes.Issue")
            else:
                python_class = get_class(python_class_name)
            # we resolve sub-objects first from the dict if deep is specified before calling deserialize on the class
            if deep:
                d = {k: deserialize(v) for k, v in d.items() if good_field(k)}
            # deserialize needs to do a shallow conversion from a dict as deep conversion is taken care of already.
            if hasattr(python_class, "from_dict"):
                return python_class.from_dict({k: v for k, v in d.items()})
            if deep:
                d = {k: deserialize(v) for k, v in d.items() if good_field(k)}
            else:
                d = {k: v for k, v in d.items() if good_field(k)}
            return python_class(**d)
        return d
    if not deep:
        return d
    if isinstance(d, str):
        return d
    if isinstance(d, bytes):
        if d.startswith(LAMBDA_START):
            return cloudpickle.loads(d[LAMBDA_START:])
        if d.startswith(FUNCTION_START):
            return cloudpickle.loads(d[FUNCTION_START:])
        if d.startswith(CLOUDPICKLE_START):
            return json.loads(d[CLOUDPICKLE_START:])
        # if d.startswith(JSON_START):
        #     return json.loads(d[JSON_START:])
        return d
    if isinstance(d, Iterable):
        return type(d)(deserialize(_) for _ in d)
    return d



import random
from typing import (
    Callable,
    List,
    Optional,
    Type,
    Union,
)


from negmas.common import AgentMechanismInterface
from negmas.outcomes import (
    Issue,
    Outcome,
)
from negmas.helpers import get_class
from .base import UtilityValue, UtilityFunction
from negmas.helpers import make_range

__all__ = [
    "LinDiscountedUFun",
    "ExpDiscountedUFun",
]


class ExpDiscountedUFun(UtilityFunction):
    """A discounted utility function based on some factor of the negotiation

    Args:
        ufun: The utility function that is being discounted
        discount: discount factor
        factor: str -> The name of the AgentMechanismInterface variable based on which discounting operate
        callable -> must receive a mechanism info object and returns a float representing the factor

    """

    def __init__(
        self,
        ufun: UtilityFunction,
        ami: "AgentMechanismInterface",
        discount: Optional[float] = None,
        factor: Union[str, Callable[["AgentMechanismInterface"], float]] = "step",
        name=None,
        reserved_value: UtilityValue = float("-inf"),
        dynamic_reservation=True,
        outcome_type: Optional[Type] = None,
        id=None,
    ):
        super().__init__(
            name=name,
            outcome_type=outcome_type,
            reserved_value=reserved_value,
            ami=ami,
            id=id,
        )
        self.ufun = ufun
        self.discount = discount
        self.factor = factor
        self.dynamic_reservation = dynamic_reservation

    @UtilityFunction.outcome_type.setter
    def outcome_type(self, value: Type):
        UtilityFunction.outcome_type.fset(self, value)
        if isinstance(self.ufun, UtilityFunction):
            self.ufun.outcome_type = value

    @UtilityFunction.ami.setter
    def ami(self, value: AgentMechanismInterface):
        UtilityFunction.ami.fset(self, value)
        if isinstance(self.ufun, UtilityFunction):
            self.ufun.ami = value

    @classmethod
    def random(
        cls,
        issues,
        reserved_value=(0.0, 1.0),
        normalized=True,
        discount_range=(0.8, 1.0),
        base_ufun_type: Union[
            str, Type[UtilityFunction]
        ] = "negmas.LinearUtilityAggregationFunction",
        **kwargs,
    ) -> "ExpDiscountedUFun":
        """Generates a random ufun of the given type"""
        reserved_value = make_range(reserved_value)
        discount_range = make_range(discount_range)
        kwargs["discount"] = (
            random.random() * (discount_range[1] - discount_range[0])
            + discount_range[0]
        )
        kwargs["reserved_value"] = (
            random.random() * (reserved_value[1] - reserved_value[0])
            + reserved_value[0]
        )
        return cls(
            get_class(base_ufun_type).random(
                issues, reserved_value=reserved_value, normalized=normalized
            ),
            **kwargs,
        )

    def eval(self, offer: "Outcome") -> UtilityValue:
        if offer is None and not self.dynamic_reservation:
            return self.reserved_value
        u = self.ufun(offer)
        if not self.discount or self.discount == 1.0 or u is None:
            return u
        if isinstance(self.factor, str):
            factor = getattr(self.ami.state, self.factor)
        else:
            factor = self.factor(self.ami.state)
        return (self.discount ** factor) * u

    def xml(self, issues: List[Issue]) -> str:
        output = self.ufun.xml(issues)
        output += "</objective>\n"
        factor = None
        if self.factor is not None:
            factor = str(self.factor)
        if self.discount is not None:
            output += f'<discount_factor value="{self.discount}" '
            if factor is not None and factor != "step":
                output += f' variable="{factor}" '
            output += "/>\n"
        return output

    def __str__(self):
        return f"{self.ufun.type}-cost:{self.discount} based on {self.factor}"

    def __getattr__(self, item):
        return getattr(self.ufun, item)

    @property
    def base_type(self):
        return self.ufun.type

    @property
    def type(self):
        return self.ufun.type + "_exponentially_discounted"


class LinDiscountedUFun(UtilityFunction):
    """A utility function with linear discounting based on some factor of the negotiation

    Args:

        ufun: The utility function that is being discounted
        cost: discount factor
        factor: str -> The name of the AgentMechanismInterface variable based on which discounting operate
        callable -> must receive a mechanism info object and returns a float representing the factor
        power: A power to raise the total cost to before discounting it from the utility_function value

    """

    def __init__(
        self,
        ufun: UtilityFunction,
        ami: "AgentMechanismInterface",
        cost: Optional[float] = None,
        factor: Union[
            str, Callable[["AgentMechanismInterface"], float]
        ] = "current_step",
        power: float = 1.0,
        name=None,
        reserved_value: UtilityValue = float("-inf"),
        dynamic_reservation=True,
        outcome_type: Optional[Type] = None,
        id=None,
    ):
        super().__init__(
            name=name,
            outcome_type=outcome_type,
            reserved_value=reserved_value,
            ami=ami,
            id=id,
        )
        self.ufun = ufun
        self.cost = cost
        self.factor = factor
        self.power = power
        self.dynamic_reservation = dynamic_reservation

    @UtilityFunction.outcome_type.setter
    def outcome_type(self, value: Type):
        UtilityFunction.outcome_type.fset(self, value)
        if isinstance(self.ufun, UtilityFunction):
            self.ufun.outcome_type = value

    @UtilityFunction.ami.setter
    def ami(self, value: AgentMechanismInterface):
        UtilityFunction.ami.fset(self, value)
        if isinstance(self.ufun, UtilityFunction):
            self.ufun.ami = value

    def eval(self, offer: "Outcome") -> UtilityValue:
        if offer is None and not self.dynamic_reservation:
            return self.reserved_value
        u = self.ufun(offer)
        if not self.cost or self.cost == 0.0:
            return u
        if isinstance(self.factor, str):
            factor = getattr(self.ami.state, self.factor)
        else:
            factor = self.factor(self.ami.state)
        return u - ((factor * self.cost) ** self.power)

    def xml(self, issues: List[Issue]) -> str:
        output = self.ufun.xml(issues)
        output += "</objective>\n"
        factor = None
        if self.factor is not None:
            factor = str(self.factor)
        if self.cost is not None:
            output += f'<cost value="{self.cost}" '
            if factor is not None and factor != "step":
                output += f' variable="{factor}" '
            if self.power is not None and self.power != 1.0:
                output += f' power="{self.power}" '
            output += "/>\n"

        return output

    def __str__(self):
        return f"{self.ufun.type}-cost:{self.cost} raised to {self.power} based on {self.factor}"

    def __getattr__(self, item):
        return getattr(self.ufun, item)

    @property
    def base_type(self):
        return self.ufun.type

    @property
    def type(self):
        return self.ufun.type + "_linearly_discounted"

    @classmethod
    def random(
        cls,
        issues,
        reserved_value=(0.0, 1.0),
        normalized=True,
        cost_range=(0.8, 1.0),
        power_range=(0.0, 1.0),
        base_ufun_type: Type[
            UtilityFunction
        ] = "negmas.LinearUtilityAggregationFunction",
        **kwargs,
    ) -> "ExpDiscountedUFun":
        """Generates a random ufun of the given type"""
        reserved_value = make_range(reserved_value)
        cost_range = make_range(cost_range)
        power_range = make_range(power_range)
        kwargs["cost"] = (
            random.random() * (cost_range[1] - cost_range[0]) + cost_range[0]
        )
        kwargs["power"] = (
            random.random() * (power_range[1] - power_range[0]) + power_range[0]
        )
        kwargs["reserved_value"] = (
            random.random() * (reserved_value[1] - reserved_value[0])
            + reserved_value[0]
        )
        return cls(
            get_class(base_ufun_type)._random(
                issues, reserved_value=kwargs["reserved_value"], normalized=normalized
            ),
            **kwargs,
        )



from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    MutableMapping,
    Optional,
    Type,
    Union,
)


from negmas.common import AgentMechanismInterface
from negmas.generics import GenericMapping, ienumerate, iget, ivalues
from negmas.helpers import Floats, gmap, ikeys
from negmas.outcomes import (
    Issue,
    OutcomeRange,
    outcome_as_tuple,
    outcome_in_range,
    Outcome,
)
from negmas.utilities.base import (
    UtilityFunction,
    UtilityValue,
    OutcomeUtilityMapping,
    OutcomeUtilityMappings,
    ExactUtilityValue,
)

__all__ = [
    "MappingUtilityFunction",
    "NonLinearUtilityAggregationFunction",
    "HyperRectangleUtilityFunction",
    "NonlinearHyperRectangleUtilityFunction",
]


class MappingUtilityFunction(UtilityFunction):
    """Outcome mapping utility function.

    This is the simplest possible utility function and it just maps a set of `Outcome`s to a set of
    `UtilityValue`(s). It is only usable with single-issue negotiations. It can be constructed with wither a mapping
    (e.g. a dict) or a callable function.

    Args:
            mapping: Either a callable or a mapping from `Outcome` to `UtilityValue`.
            default: value returned for outcomes causing exception (e.g. invalid outcomes).
            name: name of the utility function. If None a random name will be generated.
            reserved_value: The reserved value (utility of not getting an agreement = utility(None) )
            ami: an `AgentMechnismInterface` that is used mostly for setting the outcome-type in methods returning
                 outcomes.
            outcome_type: The type that should be returned from methods returning `Outcome` which can be tuple, dict or
                          any `OutcomeType`.

    Examples:

        Single issue outcome case:

        >>> issue =Issue(values=['to be', 'not to be'], name='THE problem')
        >>> print(str(issue))
        THE problem: ['to be', 'not to be']
        >>> f = MappingUtilityFunction({'to be':10.0, 'not to be':0.0})
        >>> print(list(map(f, ['to be', 'not to be'])))
        [10.0, 0.0]
        >>> f = MappingUtilityFunction(mapping={'to be':-10.0, 'not to be':10.0})
        >>> print(list(map(f, ['to be', 'not to be'])))
        [-10.0, 10.0]
        >>> f = MappingUtilityFunction(lambda x: float(len(x)))
        >>> print(list(map(f, ['to be', 'not to be'])))
        [5.0, 9.0]

        Multi issue case:

        >>> issues = [Issue((10.0, 20.0), 'price'), Issue(['delivered', 'not delivered'], 'delivery')
        ...           , Issue(5, 'quality')]
        >>> print(list(map(str, issues)))
        ['price: (10.0, 20.0)', "delivery: ['delivered', 'not delivered']", 'quality: (0, 4)']
        >>> f = MappingUtilityFunction(lambda x: x['price'] if x['delivery'] == 'delivered' else -1.0)
        >>> g = MappingUtilityFunction(lambda x: x['price'] if x['delivery'] == 'delivered' else -1.0
        ...     , default=-1000 )
        >>> f({'price': 16.0}) is None
        True
        >>> g({'price': 16.0})
        -1000
        >>> f({'price': 16.0, 'delivery':  'delivered'})
        16.0
        >>> f({'price': 16.0, 'delivery':  'not delivered'})
        -1.0

    Remarks:
        - If the mapping used failed on the outcome (for example because it is not a valid outcome), then the
        ``default`` value given to the constructor (which defaults to None) will be returned.

    """

    def __init__(
        self,
        mapping: OutcomeUtilityMapping,
        default=None,
        name: str = None,
        reserved_value: UtilityValue = float("-inf"),
        ami: AgentMechanismInterface = None,
        outcome_type: Optional[Type] = None,
        id = None,
    ) -> None:
        super().__init__(
            name=name, outcome_type=outcome_type, reserved_value=reserved_value, ami=ami, id=id,
        )
        self.mapping = mapping
        self.default = default

    def eval(self, offer: Optional[Outcome]) -> Optional[UtilityValue]:
        # noinspection PyBroadException
        if offer is None:
            return self.reserved_value
        try:
            if isinstance(offer, dict) and isinstance(self.mapping, dict):
                m = gmap(self.mapping, outcome_as_tuple(offer))
            else:
                m = gmap(self.mapping, offer)
        except Exception:
            return self.default

        return m

    def xml(self, issues: List[Issue]) -> str:
        """

        Examples:

            >>> issue =Issue(values=['to be', 'not to be'], name='THE problem')
            >>> print(str(issue))
            THE problem: ['to be', 'not to be']
            >>> f = MappingUtilityFunction({'to be':10.0, 'not to be':0.0})
            >>> print(list(map(f, ['to be', 'not to be'])))
            [10.0, 0.0]
            >>> print(f.xml([issue]))
            <issue index="1" etype="discrete" type="discrete" vtype="discrete" name="THE problem">
                <item index="1" value="to be"  cost="0"  evaluation="10.0" description="to be">
                </item>
                <item index="2" value="not to be"  cost="0"  evaluation="0.0" description="not to be">
                </item>
            </issue>
            <weight index="1" value="1.0">
            </weight>
            <BLANKLINE>
        """
        if len(issues) > 1:
            raise ValueError(
                "Cannot call xml() on a mapping utility function with more than one issue"
            )
        if issues is not None:
            issue_names = [_.name for _ in issues]
            key = issue_names[0]
        else:
            key = "0"
        output = f'<issue index="1" etype="discrete" type="discrete" vtype="discrete" name="{key}">\n'
        if isinstance(self.mapping, Callable):
            for i, k in enumerate(issues[key].all):
                if isinstance(k, tuple) or isinstance(k, list):
                    k = "-".join([str(_) for _ in k])
                output += (
                    f'    <item index="{i+1}" value="{k}"  cost="0"  evaluation="{self(k)}" description="{k}">\n'
                    f"    </item>\n"
                )
        else:
            for i, (k, v) in enumerate(ienumerate(self.mapping)):
                if isinstance(k, tuple) or isinstance(k, list):
                    k = "-".join([str(_) for _ in k])
                output += (
                    f'    <item index="{i+1}" value="{k}"  cost="0"  evaluation="{v}" description="{k}">\n'
                    f"    </item>\n"
                )
        output += "</issue>\n"
        output += '<weight index="1" value="1.0">\n</weight>\n'
        return output

    def __str__(self) -> str:
        return f"mapping: {self.mapping}\ndefault: {self.default}"

    @classmethod
    def random(
        cls,
        issues: List["Issue"],
        reserved_value=(0.0, 1.0),
        normalized=True,
        max_n_outcomes: int = 10000,
    ):
        outcomes = (
            Issue.enumerate(issues)
            if Issue.num_outcomes(issues) <= max_n_outcomes
            else Issue.sample(
                issues, max_n_outcomes, with_replacement=False, fail_if_not_enough=False
            )
        )
        return UtilityFunction.generate_random(1, outcomes)[0]


class NonLinearUtilityAggregationFunction(UtilityFunction):
    r"""A nonlinear utility function.

    Allows for the modeling of a single nonlinear utility function that combines the utilities of different issues.

    Args:
        issue_utilities: A set of mappings from issue values to utility functions. These are generic mappings so
                        \ `Callable`\ (s) and \ `Mapping`\ (s) are both accepted
        f: A nonlinear function mapping from a dict of utility_function-per-issue to a float
        name: name of the utility function. If None a random name will be generated.

    Notes:

        The utility is calculated as:

        .. math::

                u = f\\left(u_0\\left(i_0\\right), u_1\\left(i_1\\right), ..., u_n\\left(i_n\\right)\\right)

        where :math:`u_j()` is the utility function for issue :math:`j` and :math:`i_j` is value of issue :math:`j` in the
        evaluated outcome.


    Examples:
        >>> issues = [Issue((10.0, 20.0), 'price'), Issue(['delivered', 'not delivered'], 'delivery')
        ...           , Issue(5, 'quality')]
        >>> print(list(map(str, issues)))
        ['price: (10.0, 20.0)', "delivery: ['delivered', 'not delivered']", 'quality: (0, 4)']
        >>> g = NonLinearUtilityAggregationFunction({ 'price': lambda x: 2.0*x
        ...                                         , 'delivery': {'delivered': 10, 'not delivered': -10}
        ...                                         , 'quality': MappingUtilityFunction(lambda x: x-3)}
        ...         , f=lambda u: u['price']  + 2.0 * u['quality'])
        >>> float(g({'quality': 2, 'price': 14.0, 'delivery': 'delivered'})) - ((2.0*14)+2.0*(2.0-3.0))
        0.0
        >>> g = NonLinearUtilityAggregationFunction({'price'    : lambda x: 2.0*x
        ...                                         , 'delivery': {'delivered': 10, 'not delivered': -10}}
        ...         , f=lambda u: 2.0 * u['price'] )
        >>> float(g({'price': 14.0, 'delivery': 'delivered'})) - (2.0*(2.0*14))
        0.0

    """

    def xml(self, issues: List[Issue]) -> str:
        raise NotImplementedError(f"Cannot convert {self.__class__.__name__} to xml")

    def __init__(
        self,
        issue_utilities: MutableMapping[Any, GenericMapping],
        f: Callable[[Dict[Any, UtilityValue]], UtilityValue],
        name: Optional[str] = None,
        reserved_value: UtilityValue = float("-inf"),
        ami: AgentMechanismInterface = None,
        outcome_type: Optional[Type] = None,
        id = None,
    ) -> None:
        super().__init__(
            name=name, outcome_type=outcome_type, reserved_value=reserved_value, ami=ami, id=id,
        )
        self.issue_utilities = issue_utilities
        self.f = f

    def eval(self, offer: Optional["Outcome"]) -> Optional[UtilityValue]:
        if offer is None:
            return self.reserved_value
        if self.issue_utilities is None:
            raise ValueError(
                "No issue utilities were set. Call set_params() or use the constructor"
            )

        u = {}
        for k in ikeys(self.issue_utilities):
            v = iget(offer, k)
            u[k] = gmap(iget(self.issue_utilities, k), v)
        return self.f(u)

    @UtilityFunction.outcome_type.setter
    def outcome_type(self, value: AgentMechanismInterface):
        UtilityFunction.outcome_type.fset(self, value)
        if isinstance(self.issue_utilities, dict):
            for k, v in self.issue_utilities.items():
                if isinstance(v, UtilityFunction):
                    v.outcome_type = value
        else:
            for v in self.issue_utilities:
                if isinstance(v, UtilityFunction):
                    v.outcome_type = value

    @UtilityFunction.ami.setter
    def ami(self, value: AgentMechanismInterface):
        UtilityFunction.ami.fset(self, value)
        if isinstance(self.issue_utilities, dict):
            for k, v in self.issue_utilities.items():
                if isinstance(v, UtilityFunction):
                    v.ami = value
        else:
            for v in self.issue_utilities:
                if isinstance(v, UtilityFunction):
                    v.ami = value


class HyperRectangleUtilityFunction(UtilityFunction):
    """A utility function defined as a set of hyper-volumes.

    The utility function that is calulated by combining linearly a set of *probably nonlinear* functions applied in
    predefined hyper-volumes of the outcome space.

     Args:
          outcome_ranges: The outcome_ranges for which the `mappings` are defined
          weights: The *optional* weights to use for combining the outputs of the `mappings`
          ignore_issues_not_in_input: If a hyper-volumne local function is defined for some issue
          that is not in the outcome being evaluated ignore it.
          ignore_failing_range_utilities: If a hyper-volume local function fails, just assume it
          did not exist for this outcome.
          name: name of the utility function. If None a random name will be generated.

     Examples:
         We will use the following issue space of cardinality :math:`10 \times 5 \times 4`:

         >>> issues = [Issue(10), Issue(5), Issue(4)]

         Now create the utility function with

         >>> f = HyperRectangleUtilityFunction(outcome_ranges=[
         ...                                        {0: (1.0, 2.0), 1: (1.0, 2.0)},
         ...                                        {0: (1.4, 2.0), 2: (2.0, 3.0)}]
         ...                                , utilities= [2.0, lambda x: 2 * x[2] + x[0]])
         >>> g = HyperRectangleUtilityFunction(outcome_ranges=[
         ...                                        {0: (1.0, 2.0), 1: (1.0, 2.0)},
         ...                                        {0: (1.4, 2.0), 2: (2.0, 3.0)}]
         ...                                , utilities= [2.0, lambda x: 2 * x[2] + x[0]]
         ...                                , ignore_issues_not_in_input=True)
         >>> h = HyperRectangleUtilityFunction(outcome_ranges=[
         ...                                        {0: (1.0, 2.0), 1: (1.0, 2.0)},
         ...                                        {0: (1.4, 2.0), 2: (2.0, 3.0)}]
         ...                                , utilities= [2.0, lambda x: 2 * x[2] + x[0]]
         ...                                , ignore_failing_range_utilities=True)

         We can now calcualte the utility_function of some outcomes:

         * An outcome that belongs to the both outcome_ranges:
         >>> [f({0: 1.5,1: 1.5, 2: 2.5}), g({0: 1.5,1: 1.5, 2: 2.5}), h({0: 1.5,1: 1.5, 2: 2.5})]
         [8.5, 8.5, 8.5]

         * An outcome that belongs to the first hypervolume only:
         >>> [f({0: 1.5,1: 1.5, 2: 1.0}), g({0: 1.5,1: 1.5, 2: 1.0}), h({0: 1.5,1: 1.5, 2: 1.0})]
         [2.0, 2.0, 2.0]

         * An outcome that belongs to and has the first hypervolume only:
         >>> [f({0: 1.5}), g({0: 1.5}), h({0: 1.5})]
         [None, 0.0, None]

         * An outcome that belongs to the second hypervolume only:
         >>> [f({0: 1.5,2: 2.5}), g({0: 1.5,2: 2.5}), h({0: 1.5,2: 2.5})]
         [None, 6.5, None]

         * An outcome that has and belongs to the second hypervolume only:
         >>> [f({2: 2.5}), g({2: 2.5}), h({2: 2.5})]
         [None, 0.0, None]

         * An outcome that belongs to no outcome_ranges:
         >>> [f({0: 11.5,1: 11.5, 2: 12.5}), g({0: 11.5,1: 11.5, 2: 12.5}), h({0: 11.5,1: 11.5, 2: 12.5})]
         [0.0, 0.0, 0.0]


     Remarks:
         - The number of outcome_ranges, mappings, and weights must be the same
         - if no weights are given they are all assumed to equal unity
         - mappings can either by an `OutcomeUtilityMapping` or a constant.

    """

    def xml(self, issues: List[Issue]) -> str:
        """Represents the function as XML

        Args:
            issues:

        Examples:

            >>> f = HyperRectangleUtilityFunction(outcome_ranges=[
            ...                                        {0: (1.0, 2.0), 1: (1.0, 2.0)},
            ...                                        {0: (1.4, 2.0), 2: (2.0, 3.0)}]
            ...                                , utilities= [2.0, 9.0 + 4.0])
            >>> print(f.xml([Issue((0.0, 4.0), name='0'), Issue((0.0, 9.0), name='1')
            ... , Issue((0.0, 9.0), name='2')]).strip())
            <issue index="1" name="0" vtype="real" type="real" etype="real">
                <range lowerbound="0.0" upperbound="4.0"></range>
            </issue><issue index="2" name="1" vtype="real" type="real" etype="real">
                <range lowerbound="0.0" upperbound="9.0"></range>
            </issue><issue index="3" name="2" vtype="real" type="real" etype="real">
                <range lowerbound="0.0" upperbound="9.0"></range>
            </issue><utility_function maxutility="-1.0">
                <ufun type="PlainUfun" weight="1" aggregation="sum">
                    <hyperRectangle utility_function="2.0">
                        <INCLUDES index="0" min="1.0" max="2.0" />
                        <INCLUDES index="1" min="1.0" max="2.0" />
                    </hyperRectangle>
                    <hyperRectangle utility_function="13.0">
                        <INCLUDES index="0" min="1.4" max="2.0" />
                        <INCLUDES index="2" min="2.0" max="3.0" />
                    </hyperRectangle>
                </ufun>
            </utility_function>

        """
        output = ""
        for i, issue in enumerate(ivalues(issues)):
            name = issue.name
            if isinstance(issue.values, tuple):
                output += (
                    f'<issue index="{i+1}" name="{name}" vtype="real" type="real" etype="real">\n'
                    f'    <range lowerbound="{issue.values[0]}" upperbound="{issue.values[1]}"></range>\n'
                    f"</issue>"
                )
            elif isinstance(issue.values, int):
                output += (
                    f'<issue index="{i+1}" name="{name}" vtype="integer" type="integer" etype="integer" '
                    f'lowerbound="0" upperbound="{issue.values - 1}"/>\n'
                )
            else:
                output += (
                    f'<issue index="{i+1}" name="{name}" vtype="integer" type="integer" etype="integer" '
                    f'lowerbound="{min(issue.values)}" upperbound="{max(issue.values)}"/>\n'
                )
        # todo find the real maxutility
        output += '<utility_function maxutility="-1.0">\n    <ufun type="PlainUfun" weight="1" aggregation="sum">\n'
        for rect, u, w in zip(self.outcome_ranges, self.mappings, self.weights):
            output += f'        <hyperRectangle utility_function="{u * w}">\n'
            for indx in ikeys(rect):
                values = iget(rect, indx, None)
                if values is None:
                    continue
                if isinstance(values, float) or isinstance(values, int):
                    mn, mx = values, values
                elif isinstance(values, tuple):
                    mn, mx = values
                else:
                    mn, mx = min(values), max(values)
                output += (
                    f'            <INCLUDES index="{indx}" min="{mn}" max="{mx}" />\n'
                )
            output += f"        </hyperRectangle>\n"
        output += "    </ufun>\n</utility_function>"
        return output

    def __init__(
        self,
        outcome_ranges: Iterable[OutcomeRange],
        utilities: Union[Floats, OutcomeUtilityMappings],
        weights: Optional[Floats] = None,
        *,
        ignore_issues_not_in_input=False,
        ignore_failing_range_utilities=False,
        name: Optional[str] = None,
        reserved_value: UtilityValue = float("-inf"),
        ami: AgentMechanismInterface = None,
        outcome_type: Optional[Type] = None,
        id = None,
    ) -> None:
        super().__init__(
            name=name, outcome_type=outcome_type, reserved_value=reserved_value, ami=ami, id=id,
        )
        self.outcome_ranges = outcome_ranges
        self.mappings = utilities
        self.weights = weights
        self.ignore_issues_not_in_input = ignore_issues_not_in_input
        self.ignore_failing_range_utilities = ignore_failing_range_utilities
        self.adjust_params()

    def adjust_params(self):
        if self.weights is None:
            self.weights = [1.0] * len(self.outcome_ranges)

    def eval(self, offer: Optional["Outcome"]) -> Optional[UtilityValue]:
        if offer is None:
            return self.reserved_value
        u = ExactUtilityValue(0.0)
        for weight, outcome_range, mapping in zip(
            self.weights, self.outcome_ranges, self.mappings
        ):  # type: ignore
            # fail on any outcome_range that constrains issues not in the presented outcome
            if outcome_range is not None and set(ikeys(outcome_range)) - set(
                ikeys(offer)
            ) != set([]):
                if self.ignore_issues_not_in_input:
                    continue

                return None

            elif outcome_range is None or outcome_in_range(offer, outcome_range):
                if isinstance(mapping, float):
                    u += weight * mapping
                else:
                    # fail if any outcome_range utility_function cannot be calculated from the input
                    try:
                        # noinspection PyTypeChecker
                        u += weight * gmap(mapping, offer)
                    except KeyError:
                        if self.ignore_failing_range_utilities:
                            continue

                        return None

        return u


class NonlinearHyperRectangleUtilityFunction(UtilityFunction):
    """A utility function defined as a set of outcome_ranges.


    Args:
           hypervolumes: see `HyperRectangleUtilityFunction`
           mappings: see `HyperRectangleUtilityFunction`
           f: A nonlinear function to combine the results of `mappings`
           name: name of the utility function. If None a random name will be generated
    """

    def xml(self, issues: List[Issue]) -> str:
        raise NotImplementedError(f"Cannot convert {self.__class__.__name__} to xml")

    def __init__(
        self,
        hypervolumes: Iterable[OutcomeRange],
        mappings: OutcomeUtilityMappings,
        f: Callable[[List[UtilityValue]], UtilityValue],
        name: Optional[str] = None,
        reserved_value: UtilityValue = float("-inf"),
        ami: AgentMechanismInterface = None,
        outcome_type: Optional[Type] = None,
        id=None,
    ) -> None:
        super().__init__(
            name=name, outcome_type=outcome_type, reserved_value=reserved_value, ami=ami, id=id,
        )
        self.hypervolumes = hypervolumes
        self.mappings = mappings
        self.f = f

    def eval(self, offer: Optional["Outcome"]) -> Optional[UtilityValue]:
        if offer is None:
            return self.reserved_value
        if not isinstance(self.hypervolumes, Iterable):
            raise ValueError(
                "Hypervolumes are not set. Call set_params() or pass them through the constructor."
            )

        u = []
        for hypervolume, mapping in zip(self.hypervolumes, self.mappings):
            if outcome_in_range(offer, hypervolume):
                u.append(gmap(mapping, offer))
        return self.f(u)



import itertools
import random
from typing import (
    Any,
    Collection,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)


from negmas.common import AgentMechanismInterface
from negmas.generics import GenericMapping, ienumerate, iget
from negmas.helpers import gmap, ikeys
from negmas.outcomes import (
    Issue,
    Outcome,
    outcome_as,
    outcome_as_tuple,
)
from .base import UtilityFunction, UtilityValue, ExactUtilityValue
from negmas.helpers import make_range
from negmas.serialization import serialize

__all__ = [
    "LinearUtilityAggregationFunction",
    "LinearUtilityFunction",
]


class LinearUtilityFunction(UtilityFunction):
    r"""A linear utility function for multi-issue negotiations.

    Models a linear utility function using predefined weights.

    Args:
         weights: weights for combining `issue_utilities`
         name: name of the utility function. If None a random name will be generated.

    Notes:

        The utility value is calculated as:

        .. math::

            u = \sum_{i=0}^{n_{outcomes}-1} {w_i * \omega_i}


    Examples:

        >>> issues = [Issue((10.0, 20.0), 'price'), Issue(5, 'quality')]
        >>> print(list(map(str, issues)))
        ['price: (10.0, 20.0)', 'quality: (0, 4)']
        >>> f = LinearUtilityFunction({'price': 1.0, 'quality': 4.0})
        >>> float(f({'quality': 2, 'price': 14.0})
        ...       ) -  (14 + 8)
        0.0
        >>> f = LinearUtilityFunction([1.0, 2.0])
        >>> float(f((2, 14)) - (30))
        0.0

    Remarks:

        - The mapping need not use all the issues in the output as the first example shows.
        - If an outcome contains combinations of strings and numeric values that have corresponding weights, an
          exception will be raised when its utility is calculated


    """

    def __init__(
        self,
        weights: Optional[Union[Mapping[Any, float], Sequence[float]]] = None,
        biases: Optional[Union[Mapping[Any, float], Sequence[float]]] = None,
        missing_value: Optional[float] = None,
        name: Optional[str] = None,
        reserved_value: UtilityValue = float("-inf"),
        ami: AgentMechanismInterface = None,
        outcome_type: Optional[Type] = None,
        id: str = None,
    ) -> None:
        super().__init__(
            name=name,
            outcome_type=outcome_type,
            reserved_value=reserved_value,
            ami=ami,
            id=id,
        )
        self.weights = weights
        if biases is None:
            if isinstance(self.weights, dict):
                biases = dict(zip(weights.keys(), itertools.repeat(0.0)))
            else:
                biases = [0.0] * len(weights)
        self.biases = biases
        self.missing_value = missing_value

    def eval(self, offer: Optional["Outcome"]) -> Optional[UtilityValue]:
        if offer is None:
            return self.reserved_value
        # offer = outcome_for(offer, self.ami) if self.ami is not None else offer
        u = ExactUtilityValue(0.0)
        if isinstance(self.weights, dict):
            if isinstance(offer, dict):
                for k, w in self.weights.items():
                    u += w * (
                        iget(offer, k, self.missing_value) + self.biases.get(k, 0)
                    )
                return u
            else:
                if self.ami is not None:
                    newoffer = dict()
                    for i, v in enumerate(offer):
                        newoffer[self.ami.issues[i].name] = v
                elif self.issue_names is not None:
                    newoffer = dict()
                    for i, v in enumerate(offer):
                        newoffer[self.issue_names[i]] = v
                elif self.issues is not None:
                    newoffer = dict()
                    for i, v in enumerate(offer):
                        newoffer[self.issues[i].name] = v
                else:
                    raise ValueError(
                        f"Cannot find issue names but weights are given as a dict."
                    )
                for k, w in self.weights.items():
                    u += w * (
                        iget(offer, k, self.missing_value) + self.biases.get(k, 0)
                    )
                return u

        offer = outcome_as_tuple(offer)
        return sum(w * (v + b) for w, b, v in zip(self.weights, self.biases, offer))

    def xml(self, issues: List[Issue]) -> str:
        """Generates an XML string representing the utility function

        Args:
            issues:

        Examples:

            >>> issues = [Issue(values=10, name='i1'), Issue(values=4, name='i2')]
            >>> f = LinearUtilityFunction(weights=[1.0, 4.0])
            >>> print(f.xml(issues))
            <issue index="1" etype="discrete" type="discrete" vtype="discrete" name="i1">
                <item index="1" value="0.0" evaluation="0" />
                <item index="2" value="1.0" evaluation="1" />
                <item index="3" value="2.0" evaluation="2" />
                <item index="4" value="3.0" evaluation="3" />
                <item index="5" value="4.0" evaluation="4" />
                <item index="6" value="5.0" evaluation="5" />
                <item index="7" value="6.0" evaluation="6" />
                <item index="8" value="7.0" evaluation="7" />
                <item index="9" value="8.0" evaluation="8" />
                <item index="10" value="9.0" evaluation="9" />
            </issue>
            <issue index="2" etype="discrete" type="discrete" vtype="discrete" name="i2">
                <item index="1" value="0.0" evaluation="0" />
                <item index="2" value="1.0" evaluation="1" />
                <item index="3" value="2.0" evaluation="2" />
                <item index="4" value="3.0" evaluation="3" />
            </issue>
            <weight index="1" value="1.0">
            </weight>
            <weight index="2" value="4.0">
            </weight>
            <BLANKLINE>

        """
        output = ""
        keys = list(ikeys(issues))
        for i, k in enumerate(keys):
            issue = iget(issues, k)
            issue_name = issue.name
            if issue.is_float():
                output += f'<issue index="{i + 1}" etype="real" type="real" vtype="real" name="{issue_name}">\n'
                output += f'<range lowerbound = {issue.min_value} upperbound = {issue.max_value} ></range>'
            # elif issue.is_integer():
            #     output += f'<issue index="{i + 1}" etype="integer" type="integer" vtype="integer" name="{issue_name}">\n'
            #     output += f'<range lowerbound = {issue.min_value} upperbound = {issue.max_value} ></range>'
            else:
                output += f'<issue index="{i+1}" etype="discrete" type="discrete" vtype="discrete" name="{issue_name}">\n'
            vals = iget(issues, k).all
            bias = iget(self.biases, k, 0.0)
            for indx, u in enumerate(vals):
                output += f'    <item index="{indx+1}" value="{u+ bias}" evaluation="{u}" />\n'
            output += "</issue>\n"
        if isinstance(issues, dict):
            if isinstance(self.weights, dict):
                weights = self.weights
            else:
                weights = {k: v for k, v in zip(ikeys(issues), self.weights)}
        else:
            if isinstance(self.weights, list) or isinstance(self.weights, tuple):
                weights = list(self.weights)
            else:
                weights = list(self.weights.get(i.name, 1.0) for i in issues)

        for i, k in enumerate(keys):
            output += f'<weight index="{i+1}" value="{iget(weights, k)}">\n</weight>\n'
        return output

    def __str__(self):
        return f"w: {self.weights}, b: {self.biases}"

    def utility_range(
        self,
        issues: List[Issue] = None,
        outcomes: Collection[Outcome] = None,
        infeasible_cutoff: Optional[float] = None,
        return_outcomes=False,
        max_n_outcomes=1000,
        ami=None,
    ) -> Union[
        Tuple[UtilityValue, UtilityValue],
        Tuple[UtilityValue, UtilityValue, Outcome, Outcome],
    ]:
        # The minimum and maximum must be at one of the edges of the outcome space. Just enumerate them
        if issues is not None:
            ranges = [(i.min_value, i.max_value) for i in issues]
            u = sorted(
                [
                    (
                        self(
                            outcome_as(
                                outcome, self.outcome_type, [_.name for _ in issues]
                            )
                        ),
                        outcome,
                    )
                    for outcome in itertools.product(*ranges)
                ]
            )
            if return_outcomes:
                return (
                    u[0][0],
                    u[-1][0],
                    outcome_as(u[0][1], self.outcome_type, [_.name for _ in issues]),
                    outcome_as(u[-1][1], self.outcome_type, [_.name for _ in issues]),
                )
            return u[0][0], u[-1][0]
        return super().utility_range(
            issues, outcomes, infeasible_cutoff, return_outcomes, max_n_outcomes
        )

    @classmethod
    def random(
        cls, issues: List["Issue"], reserved_value=(0.0, 1.0), normalized=True, **kwargs
    ):
        from negmas.utilities.ops import normalize

        reserved_value = make_range(reserved_value)
        n_issues = len(issues)
        r = reserved_value if reserved_value is not None else random.random()
        s = 0.0
        weights = [2 * (random.random() - 0.5) for _ in range(n_issues)]
        biases = [2 * (random.random() - 0.5) for _ in range(n_issues)]
        ufun = cls(
            weights=weights,
            biases=biases,
            reserved_value=random.random() * (reserved_value[1] - reserved_value[0])
            + reserved_value[0],
        )
        if normalized:
            return normalize(
                ufun,
                outcomes=Issue.discretize_and_enumerate(
                    issues, n_discretization=10, max_n_outcomes=10000
                ),
            )
        return ufun

    def to_dict(self):
        return dict(
            weights=self.weights,
            biases=self.biases,
            missing_value=self.missing_value,
            name=self.name,
            id=self.id,
            reserved_value=self.reserved_value,
        )


def _rand_mapping(x):
    return (random.random() - 0.5) * x


def random_mapping(issue: "Issue"):
    if issubclass(issue.value_type, int) or issubclass(issue.value_type, float):
        return _rand_mapping
    return zip(issue.values, [random.random() - 0.5 for _ in range(issue.cardinality)])


class LinearUtilityAggregationFunction(UtilityFunction):
    r"""A linear aggregation utility function for multi-issue negotiations.

    Models a linear utility function using predefined weights:\.

    Args:
         issue_utilities: utility functions for individual issues
         weights: weights for combining `issue_utilities`
         name: name of the utility function. If None a random name will be generated.

    Notes:

        The utility value is calculated as:

        .. math::

            u = \sum_{i=0}^{n_{outcomes}-1} {w_i * u_i(\omega_i)}


    Examples:

        >>> from negmas.utilities.nonlinear import MappingUtilityFunction
        >>> issues = [Issue((10.0, 20.0), 'price'), Issue(['delivered', 'not delivered'], 'delivery')
        ...           , Issue(5, 'quality')]
        >>> print(list(map(str, issues)))
        ['price: (10.0, 20.0)', "delivery: ['delivered', 'not delivered']", 'quality: (0, 4)']
        >>> f = LinearUtilityAggregationFunction({'price': lambda x: 2.0*x
        ...                          , 'delivery': {'delivered': 10, 'not delivered': -10}
        ...                          , 'quality': MappingUtilityFunction(lambda x: x-3)}
        ...         , weights={'price': 1.0, 'delivery': 2.0, 'quality': 4.0})
        >>> float(f({'quality': 2, 'price': 14.0, 'delivery': 'delivered'})
        ...       ) -  (1.0*(2.0*14)+2.0*10+4.0*(2.0-3.0))
        0.0
        >>> f = LinearUtilityAggregationFunction({'price': lambda x: 2.0*x
        ...                          , 'delivery': {'delivered': 10, 'not delivered': -10}}
        ...         , weights={'price': 1.0, 'delivery': 2.0})
        >>> float(f({'quality': 2, 'price': 14.0, 'delivery': 'delivered'})) - (1.0*(2.0*14)+2.0*10)
        0.0

        You can use lists instead of dictionaries for defining outcomes, weights
        but that is less readable

        >>> f = LinearUtilityAggregationFunction([lambda x: 2.0*x
        ...                          , {'delivered': 10, 'not delivered': -10}
        ...                          , MappingUtilityFunction(lambda x: x-3)]
        ...         , weights=[1.0, 2.0, 4.0])
        >>> float(f((14.0, 'delivered', 2))) - (1.0*(2.0*14)+2.0*10+4.0*(2.0-3.0))
        0.0

    Remarks:
        The mapping need not use all the issues in the output as the last example show.

    """

    def __init__(
        self,
        issue_utilities: Union[
            MutableMapping[Any, GenericMapping], Sequence[GenericMapping]
        ],
        weights: Optional[Union[Mapping[Any, float], Sequence[float]]] = None,
        name: Optional[str] = None,
        reserved_value: UtilityValue = float("-inf"),
        ami: AgentMechanismInterface = None,
        outcome_type: Optional[Type] = None,
        id: str = None,
    ) -> None:
        from negmas.utilities.nonlinear import MappingUtilityFunction

        super().__init__(
            name=name,
            outcome_type=outcome_type,
            reserved_value=reserved_value,
            ami=ami,
            id=id,
        )
        if weights is None:
            weights = (
                {i: 1.0 for i in ikeys(issue_utilities)}
                if isinstance(issue_utilities, dict)
                else [1.0] * len(issue_utilities)
            )
        if isinstance(weights, dict) and not isinstance(issue_utilities, dict):
            raise ValueError(
                f"Type of weights is {type(weights)} but type of issue_utilities is {type(issue_utilities)}"
            )
        if not isinstance(weights, dict) and isinstance(issue_utilities, dict):
            raise ValueError(
                f"Type of weights is {type(weights)} but type of issue_utilities is {type(issue_utilities)}"
            )
        self.issue_utilities = issue_utilities
        self.weights = weights
        for k, v in ienumerate(self.issue_utilities):
            self.issue_utilities[k] = (
                v if isinstance(v, UtilityFunction) else MappingUtilityFunction(v)
            )
        if isinstance(issue_utilities, dict):
            self.issue_indices = dict(
                zip(self.issue_utilities.keys(), range(len(self.issue_utilities)))
            )
        else:
            self.issue_indices = dict(
                zip(range(len(self.issue_utilities)), range(len(self.issue_utilities)))
            )

    # @UtilityFunction.outcome_type.setter
    # def outcome_type(self, value: AgentMechanismInterface):
    #     UtilityFunction.outcome_type.fset(self, value)
    #     if isinstance(self.issue_utilities, dict):
    #         for k, v in self.issue_utilities.items():
    #             if isinstance(v, UtilityFunction):
    #                 v.outcome_type = value
    #     else:
    #         for v in self.issue_utilities:
    #             if isinstance(v, UtilityFunction):
    #                 v.outcome_type = value
    #
    # @UtilityFunction.ami.setter
    # def ami(self, value: AgentMechanismInterface):
    #     UtilityFunction.ami.fset(self, value)
    #     if isinstance(self.issue_utilities, dict):
    #         for k, v in self.issue_utilities.items():
    #             if isinstance(v, UtilityFunction):
    #                 v.ami = value
    #     else:
    #         for v in self.issue_utilities:
    #             if isinstance(v, UtilityFunction):
    #                 v.ami = value
    #
    def eval(self, offer: Optional["Outcome"]) -> Optional[UtilityValue]:
        if offer is None:
            return self.reserved_value
        u = ExactUtilityValue(0.0)
        for k in ikeys(self.issue_utilities):
            if isinstance(offer, tuple):
                v = iget(offer, self.issue_indices[k])
            else:
                v = iget(offer, k)
            current_utility = gmap(iget(self.issue_utilities, k), v)
            if current_utility is None:
                return None

            w = iget(self.weights, k)  # type: ignore
            if w is None:
                return None
            try:
                u += w * current_utility
            except FloatingPointError:
                continue
        return u

    def xml(self, issues: List[Issue]) -> str:
        """Generates an XML string representing the utility function

        Args:
            issues:

        Examples:

            >>> from negmas.utilities.nonlinear import MappingUtilityFunction
            >>> issues = [Issue(values=10, name='i1'), Issue(values=['delivered', 'not delivered'], name='i2')
            ...     , Issue(values=4, name='i3')]
            >>> f = LinearUtilityAggregationFunction([lambda x: 2.0*x
            ...                          , {'delivered': 10, 'not delivered': -10}
            ...                          , MappingUtilityFunction(lambda x: x-3)]
            ...         , weights=[1.0, 2.0, 4.0])
            >>> print(f.xml(issues))
            <issue index="1" etype="discrete" type="discrete" vtype="discrete" name="i1">
                <item index="1" value="0" evaluation="0.0" />
                <item index="2" value="1" evaluation="2.0" />
                <item index="3" value="2" evaluation="4.0" />
                <item index="4" value="3" evaluation="6.0" />
                <item index="5" value="4" evaluation="8.0" />
                <item index="6" value="5" evaluation="10.0" />
                <item index="7" value="6" evaluation="12.0" />
                <item index="8" value="7" evaluation="14.0" />
                <item index="9" value="8" evaluation="16.0" />
                <item index="10" value="9" evaluation="18.0" />
            </issue>
            <issue index="2" etype="discrete" type="discrete" vtype="discrete" name="i2">
                <item index="1" value="delivered" evaluation="10" />
                <item index="2" value="not delivered" evaluation="-10" />
            </issue>
            <issue index="3" etype="discrete" type="discrete" vtype="discrete" name="i3">
                <item index="1" value="0" evaluation="-3" />
                <item index="2" value="1" evaluation="-2" />
                <item index="3" value="2" evaluation="-1" />
                <item index="4" value="3" evaluation="0" />
            </issue>
            <weight index="1" value="1.0">
            </weight>
            <weight index="2" value="2.0">
            </weight>
            <weight index="3" value="4.0">
            </weight>
            <BLANKLINE>
            >>> print(f.xml({i:_ for i, _ in enumerate(issues)}))
            <issue index="1" etype="discrete" type="discrete" vtype="discrete" name="i1">
                <item index="1" value="0" evaluation="0.0" />
                <item index="2" value="1" evaluation="2.0" />
                <item index="3" value="2" evaluation="4.0" />
                <item index="4" value="3" evaluation="6.0" />
                <item index="5" value="4" evaluation="8.0" />
                <item index="6" value="5" evaluation="10.0" />
                <item index="7" value="6" evaluation="12.0" />
                <item index="8" value="7" evaluation="14.0" />
                <item index="9" value="8" evaluation="16.0" />
                <item index="10" value="9" evaluation="18.0" />
            </issue>
            <issue index="2" etype="discrete" type="discrete" vtype="discrete" name="i2">
                <item index="1" value="delivered" evaluation="10" />
                <item index="2" value="not delivered" evaluation="-10" />
            </issue>
            <issue index="3" etype="discrete" type="discrete" vtype="discrete" name="i3">
                <item index="1" value="0" evaluation="-3" />
                <item index="2" value="1" evaluation="-2" />
                <item index="3" value="2" evaluation="-1" />
                <item index="4" value="3" evaluation="0" />
            </issue>
            <weight index="1" value="1.0">
            </weight>
            <weight index="2" value="2.0">
            </weight>
            <weight index="3" value="4.0">
            </weight>
            <BLANKLINE>

        """
        output = ""
        keys = list(ikeys(issues))
        for i, k in enumerate(keys):
            issue = iget(issues, k)
            issue_name = issue.name
            if issue.is_float():
                output += f'<issue index="{i + 1}" etype="real" type="real" vtype="real" name="{issue_name}">\n'
                output += f'<range lowerbound = {issue.min_value} upperbound = {issue.max_value} ></range>'
            # elif issue.is_integer():
            #     output += f'<issue index="{i + 1}" etype="integer" type="integer" vtype="integer" name="{issue_name}">\n'
            #     output += f'<range lowerbound = {issue.min_value} upperbound = {issue.max_value} ></range>'
            else:
                output += f'<issue index="{i+1}" etype="discrete" type="discrete" vtype="discrete" name="{issue_name}">\n'
                vals = iget(issues, k).all
                for indx, v in enumerate(vals):
                    try:
                        u = gmap(iget(self.issue_utilities, issue_name), v)
                    except:
                        u = gmap(iget(self.issue_utilities, k), v)
                    v_ = (
                        v
                        if not (isinstance(v, tuple) or isinstance(v, list))
                        else "-".join([str(_) for _ in v])
                    )
                    output += (
                        f'    <item index="{indx+1}" value="{v_}" evaluation="{u}" />\n'
                    )
            output += "</issue>\n"
        if isinstance(issues, dict):
            if isinstance(self.weights, dict):
                weights = self.weights
            else:
                weights = {k: v for k, v in zip(ikeys(issues), self.weights)}
        else:
            if isinstance(self.weights, list):
                weights = self.weights
            else:
                weights = list(self.weights.get(i.name, 1.0) for i in issues)

        for i, k in enumerate(keys):
            output += f'<weight index="{i+1}" value="{iget(weights, k)}">\n</weight>\n'
        return output

    def __str__(self):
        return f"u: {self.issue_utilities}\n w: {self.weights}"

    def to_dict(self):
        return dict(
            weights=self.weights,
            issue_utilities=serialize(self.issue_utilities),
            name=self.name,
            id=self.id,
            reserved_value=self.reserved_value,
        )

    def utility_range(
        self,
        issues: List[Issue] = None,
        outcomes: Collection[Outcome] = None,
        infeasible_cutoff: Optional[float] = None,
        return_outcomes=False,
        max_n_outcomes=1000,
        ami=None,
    ) -> Union[
        Tuple[UtilityValue, UtilityValue],
        Tuple[UtilityValue, UtilityValue, Outcome, Outcome],
    ]:
        # The minimum and maximum must be at one of the edges of the outcome space. Just enumerate them

        # TODO test this method and add other methods for utility operations
        if issues is not None:
            uranges, vranges = [], []
            for issue in issues:
                u = self.issue_utilities[issue.name]
                mx, mn = float("-inf"), float("inf")
                mxv, mnv = None, None
                for v in issue.alli(n=max_n_outcomes):
                    uval = u(v)
                    if uval >= mx:
                        mx, mxv = uval, v
                    if uval < mn:
                        mn, mnv = uval, v
                vranges.append((mnv, mxv))
                uranges.append((mn, mx))

            best_outcome, worst_outcome = [], []
            best_util, worst_util = 0.0, 0.0
            for w, urng, vrng in zip(self.weights.values(), uranges, vranges):
                if w > 0:
                    best_util += w * urng[1]
                    best_outcome.append(vrng[1])
                    worst_util += w * urng[0]
                    worst_outcome.append(vrng[0])
                else:
                    best_util += w * urng[0]
                    best_outcome.append(vrng[0])
                    worst_util += w * urng[1]
                    worst_outcome.append(vrng[1])

            if return_outcomes:
                return (worst_util, best_util, worst_outcome, best_outcome)
            return worst_util, best_util
        return super().utility_range(
            issues, outcomes, infeasible_cutoff, return_outcomes, max_n_outcomes
        )

    @classmethod
    def random(
        cls, issues: List["Issue"], reserved_value=(0.0, 1.0), normalized=True, **kwargs
    ):
        from negmas.utilities.ops import normalize

        reserved_value = make_range(reserved_value)

        n_issues = len(issues)
        # r = reserved_value if reserved_value is not None else random.random()
        s = 0.0
        weights = dict(
            zip([_.name for _ in issues], [random.random() for _ in range(n_issues)])
        )
        s = sum(_ for _ in weights.values())
        if s:
            for k, v in weights.items():
                weights[k] = v / s
        issue_utilities = dict(
            zip([_.name for _ in issues], [random_mapping(issue) for issue in issues])
        )

        ufun = cls(
            weights=weights,
            issue_utilities=issue_utilities,
            reserved_value=random.random() * (reserved_value[1] - reserved_value[0])
            + reserved_value[0],
            **kwargs,
        )
        if normalized:
            return normalize(
                ufun,
                outcomes=Issue.discretize_and_enumerate(issues, max_n_outcomes=5000),
            )

    # def outcome_with_utility(
    #     self,
    #     rng: Tuple[Optional[float], Optional[float]],
    #     issues: List[Issue] = None,
    #     outcomes: List[Outcome] = None,
    #     n_trials: int = 100,
    # ) -> Optional["Outcome"]:
    #     """
    #     Gets one outcome within the given utility range or None on failure
    #
    #     Args:
    #         self: The utility function
    #         rng: The utility range
    #         issues: The issues the utility function is defined on
    #         outcomes: The outcomes to sample from
    #         n_trials: Not used
    #
    #     Returns:
    #
    #         - Either issues, or outcomes should be given but not both
    #
    #     """
    #



import random
from typing import (
    Callable,
    Iterable,
    List,
    Optional,
    Type,
    Any,
    Dict,
)


from negmas.common import AgentMechanismInterface
from negmas.serialization import serialize, deserialize
from negmas.outcomes import (
    Issue,
    Outcome,
)
from .base import UtilityValue, UtilityFunction, ExactUtilityValue
from negmas.helpers import make_range, get_full_type_name
from negmas.serialization import PYTHON_CLASS_IDENTIFIER

__all__ = [
    "ComplexWeightedUtilityFunction",
    "ComplexNonlinearUtilityFunction",
]


class ComplexWeightedUtilityFunction(UtilityFunction):
    """A utility function composed of linear aggregation of other utility functions

    Args:
        ufuns: An iterable of utility functions
        weights: Weights used for combination
        name: Utility function name

    """

    def __init__(
        self,
        ufuns: Iterable[UtilityFunction],
        weights: Optional[Iterable[float]] = None,
        name=None,
        reserved_value: UtilityValue = float("-inf"),
        ami: AgentMechanismInterface = None,
        outcome_type: Optional[Type] = None,
        id: str = None,
    ):
        super().__init__(
            name=name,
            outcome_type=outcome_type,
            reserved_value=reserved_value,
            ami=ami,
            id=id,
        )
        self.ufuns = list(ufuns)
        if weights is None:
            weights = [1.0] * len(self.ufuns)
        self.weights = list(weights)

    @UtilityFunction.outcome_type.setter
    def outcome_type(self, value: Type):
        UtilityFunction.outcome_type.fset(self, value)
        for ufun in self.ufuns:
            if isinstance(ufun, UtilityFunction):
                ufun.outcome_type = value

    @UtilityFunction.ami.setter
    def ami(self, value):
        UtilityFunction.ami.fset(self, value)
        for ufun in self.ufuns:
            if hasattr(ufun, "ami"):
                ufun.ami = value

    def eval(self, offer: "Outcome") -> UtilityValue:
        """Calculate the utility_function value for a given outcome.

        Args:
            offer: The offer to be evaluated.


        Remarks:
            - You cannot return None from overriden apply() functions but raise an exception (ValueError) if it was
              not possible to calculate the UtilityValue.
            - Return A UtilityValue not a float for real-valued utilities for the benefit of inspection code.

        Returns:
            UtilityValue: The utility_function value which may be a distribution. If `None` it means the utility_function value cannot be
            calculated.
        """
        if offer is None:
            return self.reserved_value
        u = ExactUtilityValue(0.0)
        failure = False
        for f, w in zip(self.ufuns, self.weights):
            util = f(offer)
            if util is not None:
                u += w * util
            else:
                failure = True
        return u if not failure else None

    def xml(self, issues: List[Issue]) -> str:
        output = ""
        # @todo implement weights. Here I assume they are always 1.0
        for f, _ in zip(self.ufuns, self.weights):
            this_output = f.xml(issues)
            if this_output:
                output += this_output
            else:
                output += str(vars(f))
        return output

    def to_dict(self) -> Dict[str, Any]:
        return {
            "ufuns": [serialize(_) for _ in self.ufuns],
            "weights": self.weights,
            "id": self.id,
            "name": self.name,
            "reserved_value": self.reserved_value,
            PYTHON_CLASS_IDENTIFIER: get_full_type_name(type(self)),
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any]):
        if PYTHON_CLASS_IDENTIFIER in d.keys():
            d = {k: v for k, v in d.items() if k != PYTHON_CLASS_IDENTIFIER}
        d["ufuns"] = [deserialize(_) for _ in d["ufuns"]]
        return cls(**d)


class ComplexNonlinearUtilityFunction(UtilityFunction):
    """A utility function composed of nonlinear aggregation of other utility functions

    Args:
        ufuns: An iterable of utility functions
        combination_function: The function used to combine results of ufuns
        name: Utility function name

    """

    def __init__(
        self,
        ufuns: Iterable[UtilityFunction],
        combination_function=Callable[[Iterable[UtilityValue]], UtilityValue],
        name=None,
        reserved_value: UtilityValue = float("-inf"),
        ami: AgentMechanismInterface = None,
        outcome_type: Optional[Type] = None,
        id: str = None,
    ):
        super().__init__(
            name=name,
            outcome_type=outcome_type,
            reserved_value=reserved_value,
            ami=ami,
            id=id,
        )
        self.ufuns = list(ufuns)
        self.combination_function = combination_function

    @UtilityFunction.outcome_type.setter
    def outcome_type(self, value: Type):
        UtilityFunction.outcome_type.fset(self, value)
        for ufun in self.ufuns:
            if hasattr(ufun, "ami"):
                ufun.outcome_type = value

    @UtilityFunction.ami.setter
    def ami(self, value):
        UtilityFunction.ami.fset(self, value)
        for ufun in self.ufuns:
            if hasattr(ufun, "ami"):
                ufun.ami = value

    def eval(self, offer: "Outcome") -> UtilityValue:
        """Calculate the utility_function value for a given outcome.

        Args:
            offer: The offer to be evaluated.


        Remarks:
            - You cannot return None from overriden apply() functions but raise an exception (ValueError) if it was
              not possible to calculate the UtilityValue.
            - Return A UtilityValue not a float for real-valued utilities for the benefit of inspection code.

        Returns:
            UtilityValue: The utility_function value which may be a distribution. If `None` it means the utility_function value cannot be
            calculated.
        """
        if offer is None:
            return self.reserved_value
        return self.combination_function([f(offer) for f in self.ufuns])

    def xml(self, issues: List[Issue]) -> str:
        raise NotImplementedError(f"Cannot convert {self.__class__.__name__} to xml")



from typing import (
    Collection,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
    TYPE_CHECKING,
)

import numpy as np

from negmas.outcomes import (
    Issue,
    Outcome,
)

from .base import UtilityFunction, UtilityValue
from negmas.common import AgentMechanismInterface

__all__ = [
    "pareto_frontier",
    "make_discounted_ufun",
    "normalize",
    "outcome_with_utility",
    "utility_range",
]


def make_discounted_ufun(
    ufun: UtilityFunction,
    ami: AgentMechanismInterface,
    cost_per_round: float = None,
    power_per_round: float = None,
    discount_per_round: float = None,
    cost_per_relative_time: float = None,
    power_per_relative_time: float = None,
    discount_per_relative_time: float = None,
    cost_per_real_time: float = None,
    power_per_real_time: float = None,
    discount_per_real_time: float = None,
    dynamic_reservation: bool = True,
):
    from negmas.utilities.discounted import LinDiscountedUFun, ExpDiscountedUFun

    if cost_per_round is not None and cost_per_round > 0.0:
        ufun = LinDiscountedUFun(
            ufun=ufun,
            ami=ami,
            cost=cost_per_round,
            factor="step",
            power=power_per_round,
            dynamic_reservation=dynamic_reservation,
        )
    if cost_per_relative_time is not None and cost_per_relative_time > 0.0:
        ufun = LinDiscountedUFun(
            ufun=ufun,
            ami=ami,
            cost=cost_per_relative_time,
            factor="relative_time",
            power=power_per_relative_time,
            dynamic_reservation=dynamic_reservation,
        )
    if cost_per_real_time is not None and cost_per_real_time > 0.0:
        ufun = LinDiscountedUFun(
            ufun=ufun,
            ami=ami,
            cost=cost_per_real_time,
            factor="real_time",
            power=power_per_real_time,
            dynamic_reservation=dynamic_reservation,
        )
    if discount_per_round is not None and discount_per_round > 0.0:
        ufun = ExpDiscountedUFun(
            ufun=ufun,
            ami=ami,
            discount=discount_per_round,
            factor="step",
            dynamic_reservation=dynamic_reservation,
        )
    if discount_per_relative_time is not None and discount_per_relative_time > 0.0:
        ufun = ExpDiscountedUFun(
            ufun=ufun,
            ami=ami,
            discount=discount_per_relative_time,
            factor="relative_time",
            dynamic_reservation=dynamic_reservation,
        )
    if discount_per_real_time is not None and discount_per_real_time > 0.0:
        ufun = ExpDiscountedUFun(
            ufun=ufun,
            ami=ami,
            discount=discount_per_real_time,
            factor="real_time",
            dynamic_reservation=dynamic_reservation,
        )
    return ufun


def _pareto_frontier(
    points, eps=-1e-18, sort_by_welfare=False
) -> Tuple[List[Tuple[float]], List[int]]:
    """Finds the pareto-frontier of a set of points

    Args:
        points: list of points
        eps: A (usually negative) small number to treat as zero during calculations
        sort_by_welfare: If True, the results are sorted descindingly by total welfare

    Returns:

    """
    points = np.asarray(points)
    n = len(points)
    indices = np.array(range(n))
    for j in range(points.shape[1]):
        order = points[:, 0].argsort()[-1::-1]
        points = points[order]
        indices = indices[order]

    frontier = [(indices[0], points[0, :])]
    for p in range(1, n):
        current = points[p, :]
        for i, (_, f) in enumerate(frontier):
            current_better, current_worse = current > f, current < f
            if np.all(current == f):
                break
            if not np.any(current_better) and np.any(current_worse):
                # current is dominated, break
                break
            if np.any(current_better):
                if not np.any(current_worse):
                    # current dominates f, append it, remove f and scan for anything else dominated by current
                    for j, (_, g) in enumerate(frontier[i + 1 :]):
                        if np.all(current == g):
                            frontier = frontier[:i] + frontier[i + 1 :]
                            break
                        if np.any(current > g) and not np.any(current < g):
                            frontier = frontier[:j] + frontier[j + 1 :]
                    else:
                        frontier[i] = (indices[p], current)
                else:
                    # neither current nor f dominate each other, append current only if it is not
                    # dominated by anything in frontier
                    for j, (_, g) in enumerate(frontier[i + 1 :]):
                        if np.all(current == g) or (
                            np.any(g > current) and not np.any(current > g)
                        ):
                            break
                    else:
                        frontier.append((indices[p], current))
    if sort_by_welfare:
        welfare = [np.sum(_[1]) for _ in frontier]
        indx = sorted(range(len(welfare)), key=lambda x: welfare[x], reverse=True)
        frontier = [frontier[_] for _ in indx]
    return [tuple(_[1]) for _ in frontier], [_[0] for _ in frontier]


def pareto_frontier(
    ufuns: Iterable[UtilityFunction],
    outcomes: Iterable[Outcome] = None,
    issues: Iterable[Issue] = None,
    n_discretization: Optional[int] = 10,
    sort_by_welfare=False,
) -> Tuple[List[Tuple[float]], List[int]]:
    """Finds all pareto-optimal outcomes in the list

    Args:

        ufuns: The utility functions
        outcomes: the outcomes to be checked. If None then all possible outcomes from the issues will be checked
        issues: The set of issues (only used when outcomes is None)
        n_discretization: The number of items to discretize each real-dimension into
        sort_by_welfare: If True, the resutls are sorted descendingly by total welfare

    Returns:
        Two lists of the same length. First list gives the utilities at pareto frontier points and second list gives their indices

    """

    ufuns = list(ufuns)
    if issues:
        issues = list(issues)
    if outcomes:
        outcomes = list(outcomes)

    # calculate all candidate outcomes
    if outcomes is None:
        if issues is None:
            return [], []
        outcomes = Issue.discretize_and_enumerate(issues, n_discretization)
        # outcomes = itertools.product(
        #     *[issue.alli(n=n_discretization) for issue in issues]
        # )
    points = [[ufun(outcome) for ufun in ufuns] for outcome in outcomes]
    return _pareto_frontier(points, sort_by_welfare=sort_by_welfare)


def normalize(
    ufun: UtilityFunction,
    outcomes: Collection[Outcome],
    rng: Tuple[float, float] = (0.0, 1.0),
    epsilon: float = 1e-6,
    infeasible_cutoff: Optional[float] = None,
    max_only: bool = False,
) -> UtilityFunction:
    """Normalizes a utility function to the range [0, 1]

    Args:
        ufun: The utility function to normalize
        outcomes: A collection of outcomes to normalize for
        rng: range to normalize to. Default is [0, 1]
        epsilon: A small number specifying the resolution
        infeasible_cutoff: A value under which any utility is considered infeasible and is not used in normalization
        max_only: If true, normalization is done by dividing by the max otherwise the range will be used.

    Returns:
        UtilityFunction: A utility function that is guaranteed to be normalized for the set of given outcomes

    """
    from negmas.utilities.complex import (
        ComplexNonlinearUtilityFunction,
        ComplexWeightedUtilityFunction,
    )
    from negmas.utilities.static import ConstUFun

    u = [ufun(o) for o in outcomes]
    u = [float(_) for _ in u if _ is not None]
    if infeasible_cutoff is not None:
        u = [_ for _ in u if _ > infeasible_cutoff]
    if len(u) == 0:
        return ufun
    mx, mn = max(u), (rng[0] if max_only else min(u))
    if abs(mx - 1.0) < epsilon and abs(mn) < epsilon:
        return ufun
    if mx == mn:
        if -epsilon <= mn <= 1 + epsilon:
            return ufun
        else:
            r = float(ufun.reserved_value) / mn if mn != 0.0 else 0.0
            if infeasible_cutoff is not None:
                return ComplexNonlinearUtilityFunction(
                    ufuns=[ufun],
                    combination_function=lambda x: infeasible_cutoff
                    if x[0] is None
                    else x[0]
                    if x[0] < infeasible_cutoff
                    else 0.5 * x[0] / mn,
                )
            else:
                return ComplexWeightedUtilityFunction(
                    ufuns=[ufun],
                    weights=[0.5 / mn],
                    name=ufun.name + "-normalized",
                    reserved_value=r,
                    ami=ufun.ami,
                )
    scale = (rng[1] - rng[0]) / (mx - mn) if not max_only else (rng[1] / mx)
    r = scale * (ufun.reserved_value - mn)
    if abs(mn - rng[0] / scale) < epsilon:
        return ufun
    if infeasible_cutoff is not None:
        return ComplexNonlinearUtilityFunction(
            ufuns=[ufun],
            combination_function=lambda x: infeasible_cutoff
            if x[0] is None
            else x[0]
            if x[0] < infeasible_cutoff
            else scale * (x[0] - mn) + rng[0],
        )
    else:
        return ComplexWeightedUtilityFunction(
            ufuns=[ufun, ConstUFun(-mn + rng[0] / scale)],
            weights=[scale, scale],
            name=ufun.name + "-normalized",
            reserved_value=r,
            ami=ufun.ami,
        )


def outcome_with_utility(
    ufun: UtilityFunction,
    rng: Tuple[float, float],
    issues: List[Issue] = None,
    outcomes: List[Outcome] = None,
    n_trials: int = 100,
) -> Optional["Outcome"]:
    """
    Gets one outcome within the given utility range or None on failure

    Args:
        ufun: The utility function
        rng: The utility range
        issues: The issues the utility function is defined on
        outcomes: The outcomes to sample from
        n_trials: The maximum number of trials

    Returns:

        - Either issues, or outcomes should be given but not both

    """
    return ufun.outcome_with_utility(rng, issues, outcomes, n_trials)


def utility_range(
    ufun: UtilityFunction,
    issues: List[Issue] = None,
    outcomes: Collection[Outcome] = None,
    infeasible_cutoff: Optional[float] = None,
    return_outcomes=False,
    max_n_outcomes=1000,
    ami: Optional["AgentMechanismInterface"] = None,
) -> Union[
    Tuple[UtilityValue, UtilityValue],
    Tuple[UtilityValue, UtilityValue, Outcome, Outcome],
]:
    """Finds the range of the given utility function for the given outcomes

    Args:
        ufun: The utility function
        issues: List of issues (optional)
        outcomes: A collection of outcomes (optional)
        infeasible_cutoff: A value under which any utility is considered infeasible and is not used in calculation
        return_outcomes: If true, returns an outcome with the min and another with the max utility
        max_n_outcomes: the maximum number of outcomes to try sampling (if sampling is used and outcomes are not
                        given)
        ami: Optional AMI to use (if not given the internal AMI can be used)
    Returns:
        Minumum utility, maximum utility (and if return_outcomes, an outcome at each)

    """
    return ufun.utility_range(
        issues, outcomes, infeasible_cutoff, return_outcomes, max_n_outcomes, ami
    )
